from __future__ import annotations

from typing import Tuple

import torch
from rich.columns import Columns
from rich.console import Console, Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.tree import Tree

from qadence.parameters import Parameter
from qadence.qubit_support import QubitSupport, QubitSupportType

from .abstract import AbstractBlock
from .primitive import ParametricBlock


class CompositeBlock(AbstractBlock):
    """Block which composes multiple blocks into one larger block (which can again be composed).
    Composite blocks are constructed via [`chain`][qadence.blocks.utils.chain],
    [`kron`][qadence.blocks.utils.kron], and [`add`][qadence.blocks.utils.add].
    """

    name = "CompositeBlock"
    blocks: Tuple[AbstractBlock, ...]

    @property
    def qubit_support(self) -> Tuple[int, ...]:
        from qadence.blocks.analog import AnalogBlock

        anablocks = filter(lambda b: isinstance(b, AnalogBlock), self.blocks)
        digiblocks = filter(lambda b: not isinstance(b, AnalogBlock), self.blocks)
        digital = sum([b.qubit_support for b in digiblocks], start=QubitSupport())
        analog = sum([b.qubit_support for b in anablocks], start=QubitSupport())
        return digital + analog

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        return torch.empty(0)

    @property
    def n_qubits(self) -> int:
        if self.qubit_support:
            return max(self.qubit_support) + 1
        else:
            return 0

    @property
    def n_supports(self) -> int:
        return len(self.qubit_support)

    @property
    def depth(self) -> int:
        return 1 + max([b.depth for b in self.blocks])

    def __iter__(self) -> CompositeBlock:
        self._iterator = iter(self.blocks)
        return self

    def __next__(self) -> AbstractBlock:
        return next(self._iterator)

    def __getitem__(self, item: int) -> AbstractBlock:
        return self.blocks[item]

    def __len__(self) -> int:
        return len(self.blocks)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            tree = Tree(self._block_title)
        else:
            tree = tree.add(self._block_title)
        for block in self.blocks:
            block.__rich_tree__(tree)
        return tree

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, type(self)):
            if len(self.blocks) != len(other.blocks):
                return False
            return self.tag == other.tag and all(
                [b0 == b1 for (b0, b1) in zip(self.blocks, other.blocks)]
            )
        return False

    def __contains__(self, other: object) -> bool:
        # Check containment by instance.
        if isinstance(other, AbstractBlock):
            for b in self.blocks:
                if isinstance(b, CompositeBlock) and other in b:
                    return True
                elif b == other:
                    return True
        elif isinstance(other, Parameter):
            for b in self.blocks:
                if isinstance(b, ParametricBlock) or isinstance(b, CompositeBlock):
                    if other in b:
                        return True
        # Check containment by type.
        elif isinstance(other, type):
            for b in self.blocks:
                if isinstance(b, CompositeBlock) and other in b:
                    return True
                elif type(b) == other:
                    return True
        else:
            raise TypeError(
                f"Can not check for containment between {type(self)} and {type(other)}."
            )
        return False

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "qubit_support": self.qubit_support,
            "tag": self.tag,
            "blocks": [b._to_dict() for b in self.blocks],
        }

    @classmethod
    def _from_dict(cls, d: dict) -> CompositeBlock:
        from qadence import blocks as qadenceblocks
        from qadence import operations
        from qadence.blocks.utils import _construct, tag

        blocks = [
            getattr(operations, b["type"])._from_dict(b)
            if hasattr(operations, b["type"])
            else getattr(qadenceblocks, b["type"])._from_dict(b)
            for b in d["blocks"]
        ]
        block = _construct(cls, blocks)  # type: ignore[arg-type]
        if d["tag"] is not None:
            tag(block, d["tag"])
        return block

    def dagger(self) -> CompositeBlock:  # type: ignore[override]
        reversed_blocks = tuple(block.dagger() for block in reversed(self.blocks))
        return self.__class__(reversed_blocks)  # type: ignore[arg-type]

    def __hash__(self) -> int:
        return hash(self._to_json())


class PutBlock(CompositeBlock):
    name = "put"

    def __init__(self, block: AbstractBlock, support: tuple):
        # np = max(support) + 1 - min(support)
        # nb = block.nqubits
        # assert np == nb, f"You are trying to put a block with {nb} qubits on {np} qubits."
        self.blocks = (block,)
        self._qubit_support = support
        super().__init__()

    @property
    def qubit_support(self) -> Tuple[int, ...]:
        return self._qubit_support

    @property
    def n_qubits(self) -> int:
        return max(self.qubit_support) + 1 - min(self.qubit_support)

    @property
    def _block_title(self) -> str:
        support = ",".join(str(i) for i in self.qubit_support)
        return f"put on ({support})"

    def __ascii__(self, console: Console) -> RenderableType:
        return self.blocks[0].__ascii__(console)

    def dagger(self) -> PutBlock:
        return PutBlock(self.blocks[0].dagger(), self.qubit_support)


class ChainBlock(CompositeBlock):
    """Chains blocks sequentially. Constructed via [`chain`][qadence.blocks.utils.chain]"""

    name = "chain"

    def __init__(self, blocks: Tuple[AbstractBlock, ...]):
        self.blocks = blocks

    def __ascii__(self, console: Console) -> RenderableType:
        # FIXME: deal with other paddings than 1
        padding = 1
        border_width = 1

        # FIXME: deal with primitive block heights other than 3
        h = 3

        def pad(b: AbstractBlock) -> Padding:
            top = (min(b.qubit_support) - min(self.qubit_support)) * h
            return Padding(b.__ascii__(console), (top, 0, 0, 0))

        cols = [pad(b) for b in self.blocks]
        w = sum([console.measure(c).minimum + padding for c in cols])
        w += padding + 2 * border_width

        return Panel(Columns(cols), title=self.tag, width=w)


class KronBlock(CompositeBlock):
    """Stacks blocks horizontally. Constructed via [`kron`][qadence.blocks.utils.kron]."""

    name = "kron"

    def __init__(self, blocks: Tuple[AbstractBlock, ...]):
        if len(blocks) == 0:
            raise NotImplementedError("Empty KronBlocks not supported")

        qubit_support = QubitSupport()
        for b in blocks:
            assert (
                QubitSupportType.GLOBAL,
            ) != b.qubit_support, "Blocks with global support cannot be kron'ed."
            assert qubit_support.is_disjoint(
                b.qubit_support
            ), "Make sure blocks act on distinct qubits!"
            qubit_support += b.qubit_support

        self.blocks = blocks

    def __ascii__(self, console: Console) -> RenderableType:
        ps = [b.__ascii__(console) for b in self.blocks]
        return Panel(Group(*ps), title=self.tag, expand=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, KronBlock):
            if len(self.blocks) != len(other.blocks):
                return False
            return self.tag == other.tag and all([b in other for b in self.blocks])
        return False


class AddBlock(CompositeBlock):
    """Adds blocks. Constructed via [`add`][qadence.blocks.utils.add]."""

    name = "add"

    def __init__(self, blocks: Tuple[AbstractBlock, ...]):
        self.blocks = blocks

    def __ascii__(self, console: Console) -> RenderableType:
        ps = [b.__ascii__(console) for b in self.blocks]
        return Panel(Group(*ps), title=self.tag, expand=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, AddBlock):
            if len(self.blocks) != len(other.blocks):
                return False
            return self.tag == other.tag and all([b in other for b in self.blocks])
        return False



from __future__ import annotations

from typing import Any, Counter, List

import numpy as np
import torch
from torch import Tensor
from torch.nn import Parameter as TorchParam

from qadence.backend import ConvertedObservable
from qadence.logger import get_logger
from qadence.measurements import Measurements
from qadence.ml_tools import promote_to_tensor
from qadence.models import QNN, QuantumModel
from qadence.utils import Endianness

logger = get_logger(__name__)


def _set_fixed_operation(
    dim: int,
    x: float | np.ndarray | Tensor | None = None,
    operation_name: str = "scale",
) -> Tensor:
    dim = dim if dim > 0 else 1
    if x is None:
        if operation_name == "shift":
            x = torch.zeros(dim)
        elif operation_name == "scale":
            x = torch.ones(dim)
        else:
            NotImplementedError
    res = promote_to_tensor(x, requires_grad=False).squeeze(0)
    assert (
        res.numel() == dim
    ), f"Number of {operation_name} values is {res.numel()}\
    and does not match number of dimensions = {dim}."
    return res


class TransformedModule(torch.nn.Module):
    """
    This class accepts a torch.nn.Module or a QuantumModel/QNN and wraps it with
    either non-trainble or trainable scaling and shifting parameters for both input and output.
    When given a torch.nn.Module, in_features and out_features need to be passed.

    Args:
        model: The original model to transform.
        in_features: The number of input dimensions of the model.
        out_features: The number of output dimensions of the model.
        input_scaling: The rescaling factor for the model input. Defaults to None.
        input_shifting: The translation factor for the model input. Defaults to None.
        output_scaling: The rescaling factor for the model output. Defaults to None.
        output_shifting: The translation factor for the model output. Defaults to None.

    Example:
    ```
    import torch
    from torch.nn import Parameter as TorchParam
    from qadence.models import QNN, TransformedModule
    from qadence.circuit import QuantumCircuit
    from qadence.blocks import chain
    from qadence.constructors import hamiltonian_factory, hea
    from qadence import Parameter, QuantumCircuit, Z

    n_qubits = 2
    phi = Parameter("phi", trainable=False)
    fm = chain(*[RY(i, phi) for i in range(n_qubits)])
    ansatz = hea(n_qubits=n_qubits, depth=3)
    observable = hamiltonian_factory(n_qubits, detuning = Z)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)

    model = QNN(circuit, observable, backend="pyqtorch", diff_mode="ad")
    batch_size = 1
    input_values = {"phi": torch.rand(batch_size, requires_grad=True)}
    pred = model(input_values)
    assert not torch.isnan(pred)

    transformed_model = TransformedModule(
        model=model,
        in_features=None,
        out_features=None,
        input_scaling=TorchParam(torch.tensor(1.0)),
        input_shifting=0.0,
        output_scaling=1.0,
        output_shifting=TorchParam(torch.tensor(0.0))
    )
    pred_transformed = transformed_model(input_values)
    ```
    """

    def __init__(
        self,
        model: torch.nn.Module | QuantumModel | QNN,
        in_features: int | None = None,
        out_features: int | None = None,
        input_scaling: TorchParam | float | int | torch.Tensor | None = None,
        input_shifting: TorchParam | float | int | torch.Tensor | None = None,
        output_scaling: TorchParam | float | int | torch.Tensor | None = None,
        output_shifting: TorchParam | float | int | torch.Tensor | None = None,
    ) -> None:
        super().__init__()
        self.model = model
        if in_features is None and out_features is None:
            assert isinstance(model, (QuantumModel, QNN))
            self.in_features = model.in_features
            self.out_features = model.out_features if model.out_features else 1
        else:
            self.in_features = in_features  # type: ignore[assignment]
            self.out_features = out_features  # type: ignore[assignment]
        if not isinstance(input_scaling, torch.Tensor):
            self.register_buffer(
                "_input_scaling",
                _set_fixed_operation(self.in_features, input_scaling, "scale"),
            )
        else:
            self._input_scaling = input_scaling
        if not isinstance(input_shifting, torch.Tensor):
            self.register_buffer(
                "_input_shifting",
                _set_fixed_operation(self.in_features, input_shifting, "shift"),
            )
        else:
            self._input_shifting = input_shifting
        if not isinstance(output_scaling, torch.Tensor):
            self.register_buffer(
                "_output_scaling",
                _set_fixed_operation(self.out_features, output_scaling, "scale"),
            )
        else:
            self._output_scaling = output_scaling
        if not isinstance(output_shifting, torch.Tensor):
            self.register_buffer(
                "_output_shifting",
                _set_fixed_operation(self.out_features, output_shifting, "shift"),
            )
        else:
            self._output_shifting = output_shifting

    def _format_to_dict(self, values: Tensor) -> dict[str, Tensor]:
        """Format an input tensor into the format required by the forward pass

        The tensor is assumed to have dimensions: n_batches x in_features where in_features
        corresponds to the number of input features of the QNN
        """

        if len(values.size()) == 1:
            values = values.reshape(-1, 1)
        if len(values.size()) != 2 or values.shape[1] != len(self.model.inputs):
            raise ValueError(
                f"Model expects in_features={self.model.in_features} but got {values.size()[1]}."
            )
        names = [p.name for p in self.model.inputs]
        res = {}
        for i, name in enumerate(names):
            res[name] = values[:, i]
        return res

    def _transform_x(self, x: dict[str, torch.Tensor] | Tensor) -> dict[str, Tensor] | Tensor:
        """
        x can either be a torch Tensor in when using torch.nn.Module, or a standard values dict.
        Scales and shifts the tensors in the values dict, containing Featureparameters.
        Transformation of inputs can be used to speed up training and avoid potential issues
        with numerical stability that can arise due to differing feature scales.
        If none are provided, it uses 0. for shifting and 1. for scaling (hence, identity).

        Arguments:
            values: A torch Tensor or a dict containing values for Featureparameters.

        Returns:
            A Tensor or dict containing transformed (scaled and/or shifted) Featureparameters.

        """

        if isinstance(self.model, (QuantumModel, QNN)):
            if not isinstance(x, dict):
                x = self._format_to_dict(x)
            return {
                key: self._input_scaling * (val + self._input_shifting) for key, val in x.items()
            }

        else:
            assert isinstance(self.model, torch.nn.Module) and isinstance(x, Tensor)
            return self._input_scaling * (x + self._input_shifting)

    def forward(self, x: dict[str, Tensor] | Tensor, *args: Any, **kwargs: Any) -> Tensor:
        y = self.model(self._transform_x(x), *args, **kwargs)
        return self._output_scaling * y + self._output_shifting

    def run(
        self,
        values: dict[str, torch.Tensor],
        state: torch.Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        return self.model.run(values=self._transform_x(values), state=state, endianness=endianness)

    def sample(
        self,
        values: dict[str, torch.Tensor],
        n_shots: int = 1000,
        state: torch.Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        return self.model.sample(  # type: ignore[no-any-return]
            values=self._transform_x(values),
            n_shots=n_shots,
            state=state,
            endianness=endianness,
        )

    def expectation(
        self,
        values: dict[str, torch.Tensor],
        observable: List[ConvertedObservable] | ConvertedObservable | None = None,
        state: torch.Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """
        Computes standard expectation, however scales and shifts the output tensor
        of the underlying model. If none are provided, it uses 0. for shifting and 1. for scaling.
        Transformation of ouputs can be used if the magnitude
        of the targets exceeds the domain (-1,1).

        """
        exp = self.model.expectation(
            values=self._transform_x(values),
            observable=observable if observable is not None else self.model._observable,
            state=state,
            protocol=protocol,
            endianness=endianness,
        )
        return self._output_scaling * exp + self._output_shifting

    def _to_dict(self, save_params: bool = True) -> dict:
        from qadence.serialization import serialize

        def store_fn(x: torch.Tensor) -> list[float]:
            res: list[float]
            if x.requires_grad:
                res = x.detach().numpy().tolist()
            else:
                res = x.numpy().tolist()
            return res  # type: ignore[no-any-return]

        _d = serialize(self.model, save_params=save_params)

        return {
            self.__class__.__name__: _d,
            "in_features": self.in_features,
            "out_features": self.out_features,
            "_input_scaling": store_fn(self._input_scaling),
            "_output_scaling": store_fn(self._output_scaling),
            "_input_shifting": store_fn(self._input_shifting),
            "_output_shifting": store_fn(self._output_shifting),
        }

    @classmethod
    def _from_dict(cls, d: dict, as_torch: bool = False) -> TransformedModule:
        from qadence.serialization import deserialize

        _m: QuantumModel | QNN = deserialize(d[cls.__name__], as_torch)  # type: ignore[assignment]
        return cls(
            _m,
            in_features=d["in_features"],
            out_features=d["out_features"],
            input_scaling=torch.tensor(d["_input_scaling"]),
            output_scaling=torch.tensor(d["_output_scaling"]),
            input_shifting=torch.tensor(d["_input_shifting"]),
            output_shifting=torch.tensor(d["_output_shifting"]),
        )



from __future__ import annotations

import os
from collections import Counter, OrderedDict
from dataclasses import asdict
from pathlib import Path
from typing import Any, Callable, Optional, Sequence

import torch
from torch import Tensor, nn

from qadence.backend import (
    Backend,
    BackendConfiguration,
    BackendName,
    ConvertedCircuit,
    ConvertedObservable,
)
from qadence.backends import backend_factory, config_factory
from qadence.backends.pytorch_wrapper import DiffMode
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.logger import get_logger
from qadence.measurements import Measurements
from qadence.utils import Endianness

logger = get_logger(__name__)


class QuantumModel(nn.Module):
    """The central class of qadence that executes `QuantumCircuit`s and make them differentiable.

    This class should be used as base class for any new quantum model supported in the qadence
    framework for information on the implementation of custom models see
    [here](/advanced_tutorials/custom-models.md).
    """

    backend: Backend
    embedding_fn: Callable
    _params: nn.ParameterDict
    _circuit: ConvertedCircuit
    _observable: list[ConvertedObservable] | None

    def __init__(
        self,
        circuit: QuantumCircuit,
        observable: list[AbstractBlock] | AbstractBlock | None = None,
        backend: BackendName | str = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        protocol: Measurements | None = None,
        configuration: BackendConfiguration | dict | None = None,
    ):
        """Initialize a generic QuantumModel instance.

        Arguments:
            circuit: The circuit that is executed.
            observable: Optional observable(s) that are used only in the `expectation` method. You
                can also provide observables on the fly to the expectation call directly.
            backend: A backend for circuit execution.
            diff_mode: A differentiability mode. Parameter shift based modes work on all backends.
                AD based modes only on PyTorch based backends.
            protocol: Optional measurement protocol. If None, use
                exact expectation value with a statevector simulator.
            configuration: Configuration for the backend.

        Raises:
            ValueError: if the `diff_mode` argument is set to None
        """
        super().__init__()

        if not isinstance(circuit, QuantumCircuit):
            TypeError(
                f"The circuit should be of type '<class QuantumCircuit>'. Got {type(circuit)}."
            )

        self.inputs = [p for p in circuit.unique_parameters if not p.trainable and not p.is_number]
        if diff_mode is None:
            raise ValueError("`diff_mode` cannot be `None` in a `QuantumModel`.")

        self.backend = backend_factory(
            backend=backend, diff_mode=diff_mode, configuration=configuration
        )

        if isinstance(observable, list) or observable is None:
            observable = observable
        else:
            observable = [observable]

        conv = self.backend.convert(circuit, observable)
        self.embedding_fn = conv.embedding_fn
        self._circuit = conv.circuit
        self._observable = conv.observable
        self._backend_name = backend
        self._diff_mode = diff_mode
        self._protocol = protocol

        self._params = nn.ParameterDict(
            {
                str(key): nn.Parameter(val, requires_grad=val.requires_grad)
                for key, val in conv.params.items()
            }
        )

    @property
    def vparams(self) -> OrderedDict:
        return OrderedDict({k: v.data for k, v in self._params.items() if v.requires_grad})

    @property
    def vals_vparams(self) -> Tensor:
        """Dictionary with parameters which are actually updated during optimization"""
        vals = torch.tensor([v for v in self._params.values() if v.requires_grad])
        vals.requires_grad = False
        return vals.flatten()

    @property
    def in_features(self) -> int:
        """Number of inputs."""
        return len(self.inputs)

    @property
    def out_features(self) -> int | None:
        """Number of outputs"""
        return 0 if self._observable is None else len(self._observable)

    @property
    def num_vparams(self) -> int:
        """The number of variational parameters"""
        return len(self.vals_vparams)

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        return self.backend.circuit(circuit)

    def observable(self, observable: AbstractBlock, n_qubits: int) -> Any:
        return self.backend.observable(observable, n_qubits)

    def reset_vparams(self, values: Sequence) -> None:
        """Reset all the variational parameters with a given list of values"""
        current_vparams = OrderedDict({k: v for k, v in self._params.items() if v.requires_grad})

        assert (
            len(values) == self.num_vparams
        ), "Pass an iterable with the values of all variational parameters"
        for i, k in enumerate(current_vparams.keys()):
            current_vparams[k].data = torch.tensor([values[i]])

    def forward(self, *args: Any, **kwargs: Any) -> Tensor:
        return self.run(*args, **kwargs)

    def run(
        self,
        values: dict[str, Tensor] = None,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        if values is None:
            values = {}
        params = self.embedding_fn(self._params, values)
        return self.backend.run(self._circuit, params, state=state, endianness=endianness)

    def sample(
        self,
        values: dict[str, torch.Tensor] = {},
        n_shots: int = 1000,
        state: torch.Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        params = self.embedding_fn(self._params, values)
        return self.backend.sample(
            self._circuit, params, n_shots=n_shots, state=state, endianness=endianness
        )

    def expectation(
        self,
        values: dict[str, Tensor] = {},
        observable: list[ConvertedObservable] | ConvertedObservable | None = None,
        state: Optional[Tensor] = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """Compute expectation using the given backend.

        Returns:
            A torch.Tensor of shape n_batches x n_obs
        """
        if observable is None:
            if self._observable is None:
                raise ValueError(
                    "Provide an AbstractBlock as the observable to compute expectation."
                    "Either pass a 'native_observable' directly to 'QuantumModel.expectation'"
                    "or pass a (non-native) '<class AbstractBlock>' to the 'QuantumModel.__init__'."
                )
            observable = self._observable

        params = self.embedding_fn(self._params, values)
        if protocol is None:
            protocol = self._protocol

        return self.backend.expectation(
            circuit=self._circuit,
            observable=observable,
            param_values=params,
            state=state,
            protocol=protocol,
            endianness=endianness,
        )

    def overlap(self) -> Tensor:
        raise NotImplementedError("The overlap method is not implemented for this model.")

    def _to_dict(self, save_params: bool = False) -> dict[str, Any]:
        if isinstance(self._observable, list):
            abs_obs = [obs.abstract._to_dict() for obs in self._observable]
        else:
            abs_obs = [dict()]

        d = {
            "circuit": self._circuit.abstract._to_dict(),
            "observable": abs_obs,
            "backend": self._backend_name,
            "diff_mode": self._diff_mode,
            "protocol": self._protocol._to_dict() if self._protocol is not None else {},
            "backend_configuration": asdict(self.backend.backend.config),  # type: ignore
        }
        param_dict_conv = {}
        if save_params:
            param_dict_conv = {name: param.data for name, param in self._params.items()}
        return {self.__class__.__name__: d, "param_dict": param_dict_conv}

    @classmethod
    def _from_dict(cls, d: dict, as_torch: bool = False) -> QuantumModel:
        from qadence.serialization import deserialize

        qm_dict = d[cls.__name__]
        qm = cls(
            circuit=QuantumCircuit._from_dict(qm_dict["circuit"]),
            observable=(
                None
                if not isinstance(qm_dict["observable"], list)
                else [deserialize(q_obs) for q_obs in qm_dict["observable"]]  # type: ignore[misc]
            ),
            backend=qm_dict["backend"],
            diff_mode=qm_dict["diff_mode"],
            protocol=Measurements._from_dict(qm_dict["protocol"]),
            configuration=config_factory(qm_dict["backend"], qm_dict["backend_configuration"]),
        )

        if as_torch:
            conv_pd = torch.nn.ParameterDict()
            param_dict = d["param_dict"]
            for n, param in param_dict.items():
                conv_pd[n] = torch.nn.Parameter(param)
            qm._params = conv_pd
        return qm

    def save(
        self, folder: str | Path, file_name: str = "quantum_model.pt", save_params: bool = True
    ) -> None:
        if not os.path.isdir(folder):
            raise FileNotFoundError
        try:
            torch.save(self._to_dict(save_params), folder / Path(file_name))
        except Exception as e:
            print(f"Unable to write QuantumModel to disk due to {e}")

    @classmethod
    def load(
        cls, file_path: str | Path, as_torch: bool = False, map_location: str | torch.device = "cpu"
    ) -> QuantumModel:
        qm_pt = {}
        if isinstance(file_path, str):
            file_path = Path(file_path)
        if os.path.isdir(file_path):
            from qadence.ml_tools.saveload import get_latest_checkpoint_name

            file_path = file_path / get_latest_checkpoint_name(file_path, "model")

        try:
            qm_pt = torch.load(file_path, map_location=map_location)
        except Exception as e:
            print(f"Unable to load QuantumModel due to {e}")
        return cls._from_dict(qm_pt, as_torch)

    def assign_parameters(self, values: dict[str, Tensor]) -> Any:
        """Return the final, assigned circuit that is used in e.g. `backend.run`"""
        params = self.embedding_fn(self._params, values)
        return self.backend.assign_parameters(self._circuit, params)



from __future__ import annotations

from itertools import chain as flatten
from typing import Callable, Dict, List

from braket.circuits.gates import (
    CZ,
    CNot,
    CPhaseShift,
    CSwap,
    H,
    I,
    Rx,
    Ry,
    Rz,
    S,
    Swap,
    T,
    X,
    Y,
    Z,
)
from braket.circuits.instruction import Instruction
from braket.parametric import FreeParameter

from qadence.blocks import AbstractBlock, CompositeBlock, PrimitiveBlock
from qadence.errors import NotSupportedError
from qadence.operations import OpName
from qadence.parameters import evaluate

single_qubit: Dict[str, Callable] = {
    OpName.I: I.i,
    OpName.H: H.h,
    OpName.X: X.x,
    OpName.Y: Y.y,
    OpName.Z: Z.z,
    OpName.S: S.s,
    OpName.T: T.t,
}
single_qubit_parameterized: Dict[str, Callable] = {
    OpName.RX: Rx.rx,
    OpName.RY: Ry.ry,
    OpName.RZ: Rz.rz,
}
two_qubit: Dict[str, Callable] = {OpName.CNOT: CNot.cnot, OpName.SWAP: Swap.swap, OpName.CZ: CZ.cz}
three_qubit: Dict[str, Callable] = {OpName.CSWAP: CSwap.cswap}
two_qubit_parametrized: Dict[str, Callable] = {
    OpName.CPHASE: CPhaseShift.cphaseshift,
}

ops_map = {
    **single_qubit,
    **single_qubit_parameterized,
    **two_qubit,
    **two_qubit_parametrized,
    **three_qubit,
}

supported_gates = list(ops_map.keys())


def BraketOperation(block: PrimitiveBlock) -> Instruction:
    operation = block.name

    if operation in single_qubit:
        return single_qubit[operation](target=block.qubit_support)

    elif operation in single_qubit_parameterized:
        ((uuid, expr),) = block.parameters.items()  # type: ignore [attr-defined]
        if expr.is_number:
            return single_qubit_parameterized[operation](
                target=block.qubit_support, angle=evaluate(expr)  # type: ignore
            )
        else:
            return single_qubit_parameterized[operation](
                target=block.qubit_support,
                angle=FreeParameter(uuid),  # type: ignore
            )

    elif operation in two_qubit:
        return two_qubit[operation](block.qubit_support[0], block.qubit_support[1])

    elif operation in two_qubit_parametrized:
        (expr,) = block.parameters.expressions()  # type: ignore [attr-defined]
        angle_value = evaluate(expr)
        return two_qubit_parametrized[operation](
            control=block.qubit_support[0],
            target=block.qubit_support[1],
            angle=angle_value,
        )
    elif operation in three_qubit:
        return three_qubit[operation](
            block.qubit_support[0], block.qubit_support[1], block.qubit_support[2]
        )
    else:
        raise NotSupportedError(
            "Operation type {} is not supported for Braket backend.".format(type(block))
        )


def convert_block(block: AbstractBlock) -> List[Instruction]:
    if isinstance(block, PrimitiveBlock):
        ops = [BraketOperation(block=block)]
    elif isinstance(block, CompositeBlock):
        ops = list(flatten(convert_block(b) for b in block.blocks))
    else:
        raise NotSupportedError(
            "Operation type {} is not supported for Braket backend.".format(type(block))
        )
    return ops



from __future__ import annotations

from functools import reduce, singledispatch

from openfermion import QubitOperator
from openfermion.utils import count_qubits

from qadence import operations
from qadence.blocks import AbstractBlock, AddBlock, CompositeBlock, PrimitiveBlock, ScaleBlock
from qadence.blocks.utils import add, kron
from qadence.parameters import evaluate


@singledispatch
def to_openfermion(block: AbstractBlock) -> QubitOperator:
    raise ValueError(f"Unable to convert type {type(block)} to QubitOperator.")


@to_openfermion.register
def _(block: PrimitiveBlock) -> QubitOperator:
    pauli, qubit = block.name, block.qubit_support[0]
    return QubitOperator(f"{pauli}{qubit}")


@to_openfermion.register
def _(block: operations.I) -> QubitOperator:
    return QubitOperator("")


@to_openfermion.register
def _(block: CompositeBlock) -> QubitOperator:
    return reduce(lambda x, y: x * y, [to_openfermion(b) for b in block.blocks])


@to_openfermion.register
def _(block: AddBlock) -> QubitOperator:
    return reduce(lambda x, y: x + y, [to_openfermion(b) for b in block.blocks])


@to_openfermion.register
def _(block: ScaleBlock) -> QubitOperator:
    op = to_openfermion(block.block)
    return op * evaluate(block.parameters.parameter)


@to_openfermion.register
def _(block: AbstractBlock) -> QubitOperator:
    return to_openfermion(block)


def from_openfermion(op: QubitOperator) -> AbstractBlock:
    n_qubits = count_qubits(op)

    def _convert_gate(gate: tuple[int, str]) -> PrimitiveBlock:
        (i, pauli) = gate
        return getattr(operations, pauli)(i)  # type: ignore [no-any-return]

    @singledispatch
    def _convert(op: QubitOperator) -> AbstractBlock:
        if isinstance(op, QubitOperator):
            return _convert(op.terms)
        else:
            raise ValueError(f"Can only conver QubitOperators. Found {type(op)}.")

    @_convert.register
    def _(op: dict) -> AbstractBlock:
        bs = [_convert(term) * coef for term, coef in op.items()]
        return add(*bs) if len(bs) > 1 else bs[0]  # type: ignore [no-any-return]

    @_convert.register
    def _(op: tuple) -> AbstractBlock:
        if len(op) == 0:
            return operations.I(n_qubits - 1)
        bs = [_convert_gate(gate) for gate in op]
        return kron(*bs) if len(bs) > 1 else bs[0]  # type: ignore [no-any-return]

    return _convert(op)



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from math import prod
from typing import Any

import pyqtorch.modules as pyq
import torch
from torch import Tensor

from qadence.backend import Backend as BackendInterface
from qadence.backend import BackendName, ConvertedCircuit, ConvertedObservable
from qadence.backends.utils import to_list_of_dicts
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.overlap import overlap_exact
from qadence.states import zero_state
from qadence.transpile import (
    add_interaction,
    blockfn_to_circfn,
    chain_single_qubit_ops,
    flatten,
    scale_primitive_blocks_only,
    transpile,
)
from qadence.utils import Endianness, int_to_basis

from .config import Configuration
from .convert_ops import convert_block, convert_observable


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    """PyQTorch backend."""

    # set standard interface parameters
    name: BackendName = BackendName.PYQTORCH
    supports_ad: bool = True
    support_bp: bool = True
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = Configuration()

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        transpilations = [
            lambda circ: add_interaction(circ, interaction=self.config.interaction),
            lambda circ: blockfn_to_circfn(chain_single_qubit_ops)(circ)
            if self.config.use_single_qubit_composition
            else blockfn_to_circfn(flatten)(circ),
            blockfn_to_circfn(scale_primitive_blocks_only),
        ]

        abstract = transpile(*transpilations)(circuit)  # type: ignore[call-overload]
        ops = convert_block(abstract.block, n_qubits=circuit.n_qubits, config=self.config)
        native = pyq.QuantumCircuit(abstract.n_qubits, ops)
        return ConvertedCircuit(native=native, abstract=abstract, original=circuit)

    def observable(self, observable: AbstractBlock, n_qubits: int) -> ConvertedObservable:
        # make sure only leaves, i.e. primitive blocks are scaled
        transpilations = [
            lambda block: chain_single_qubit_ops(block)
            if self.config.use_single_qubit_composition
            else flatten(block),
            scale_primitive_blocks_only,
        ]
        block = transpile(*transpilations)(observable)  # type: ignore[call-overload]

        (native,) = convert_observable(block, n_qubits=n_qubits, config=self.config)
        return ConvertedObservable(native=native, abstract=block, original=observable)

    def run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
        pyqify_state: bool = True,
        unpyqify_state: bool = True,
    ) -> Tensor:
        n_qubits = circuit.abstract.n_qubits

        if state is not None:
            if pyqify_state:
                if (state.ndim != 2) or (state.size(1) != 2**n_qubits):
                    raise ValueError(
                        "The initial state must be composed of tensors of size "
                        f"(batch_size, 2**n_qubits). Found: {state.size() = }."
                    )

                # PyQ expects a column vector for the initial state
                # where each element is of dim=2.
                state = state.T.reshape([2] * n_qubits + [state.size(0)])
            else:
                if prod(state.size()[:-1]) != 2**n_qubits:
                    raise ValueError(
                        "A pyqified initial state must be composed of tensors of size "
                        f"(2, 2, ..., batch_size). Found: {state.size() = }."
                    )
        else:
            # infer batch_size without state
            if len(param_values) == 0:
                batch_size = 1
            else:
                batch_size = max([len(tensor) for tensor in param_values.values()])
            state = circuit.native.init_state(batch_size=batch_size)
        state = circuit.native(state, param_values)

        # make sure that the batch dimension is the first one, as standard
        # for PyTorch, and not the last one as done in PyQ
        if unpyqify_state:
            state = torch.flatten(state, start_dim=0, end_dim=-2).t()

        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            state = invert_endianness(state)
        return state

    def _batched_expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        state = self.run(
            circuit,
            param_values=param_values,
            state=state,
            endianness=endianness,
            pyqify_state=True,
            # we are calling  the native observable directly, so we want to use pyq shapes
            unpyqify_state=False,
        )
        observable = observable if isinstance(observable, list) else [observable]
        _expectation = torch.hstack(
            [obs.native(state, param_values).reshape(-1, 1) for obs in observable]
        )
        return _expectation

    def _looped_expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        state = zero_state(circuit.abstract.n_qubits, batch_size=1) if state is None else state
        if state.size(0) != 1:
            raise ValueError(
                "Looping expectation does not make sense with batched initial state. "
                "Define your initial state with `batch_size=1`"
            )

        list_expvals = []
        observables = observable if isinstance(observable, list) else [observable]
        for vals in to_list_of_dicts(param_values):
            wf = self.run(circuit, vals, state, endianness, pyqify_state=True, unpyqify_state=False)
            exs = torch.cat([obs.native(wf, vals) for obs in observables], 0)
            list_expvals.append(exs)

        batch_expvals = torch.vstack(list_expvals)
        return batch_expvals if len(batch_expvals.shape) > 0 else batch_expvals.reshape(1)

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        fn = self._looped_expectation if self.config.loop_expectation else self._batched_expectation
        return fn(
            circuit=circuit,
            observable=observable,
            param_values=param_values,
            state=state,
            protocol=protocol,
            endianness=endianness,
        )

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        def _sample(_probs: Tensor, n_shots: int, endianness: Endianness, n_qubits: int) -> Counter:
            return Counter(
                {
                    int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
                    for k, count in enumerate(
                        torch.bincount(
                            torch.multinomial(input=_probs, num_samples=n_shots, replacement=True)
                        )
                    )
                    if count > 0
                }
            )

        wf = self.run(circuit=circuit, param_values=param_values, state=state)
        probs = torch.abs(torch.pow(wf, 2))
        return list(
            map(
                lambda _probs: _sample(
                    _probs=_probs,
                    n_shots=n_shots,
                    endianness=endianness,
                    n_qubits=circuit.abstract.n_qubits,
                ),
                probs,
            )
        )

    def assign_parameters(self, circuit: ConvertedCircuit, param_values: dict[str, Tensor]) -> Any:
        raise NotImplementedError

    @staticmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        return overlap_exact(bras, kets)

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass, fields
from typing import Any, Callable, Iterator, Tuple

from openfermion import QubitOperator
from torch import Tensor
from torch.nn import Module

from qadence.blocks import (
    AbstractBlock,
    CompositeBlock,
    ParametricBlock,
    PrimitiveBlock,
    ScaleBlock,
    TimeEvolutionBlock,
    embedding,
)
from qadence.blocks.analog import ConstantAnalogRotation, WaitBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.parameters import stringify
from qadence.types import BackendName, DiffMode, Endianness


@dataclass
class BackendConfiguration:
    _use_gate_params: bool = True
    use_sparse_observable: bool = False
    use_gradient_checkpointing: bool = False
    use_single_qubit_composition: bool = False

    def available_options(self) -> str:
        """Return as a string the available fields with types of the configuration

        Returns:
            str: a string with all the available fields, one per line
        """
        conf_msg = ""
        for field in fields(self):
            if not field.name.startswith("_"):
                conf_msg += (
                    f"Name: {field.name} - Type: {field.type} - Default value: {field.default}\n"
                )
        return conf_msg

    @classmethod
    def from_dict(cls, values: dict) -> BackendConfiguration:
        field_names = {field.name for field in fields(cls)}
        init_data = {}

        for key, value in values.items():
            if key not in field_names:
                raise ValueError(f"Unknown field in the configuration: '{key}'.")
            else:
                init_data[key] = value

        return cls(**init_data)

    def get_param_name(self, blk: AbstractBlock) -> Tuple[str, ...]:
        """Return parameter names for the current backend. Depending on which backend is in use this
        function returns either UUIDs or expressions of parameters."""
        param_ids: Tuple
        # FIXME: better type hiearchy?
        types = (TimeEvolutionBlock, ParametricBlock, ConstantAnalogRotation, WaitBlock)
        if not isinstance(blk, types):
            raise TypeError(f"Can not infer param name from {type(blk)}")
        else:
            if self._use_gate_params:
                param_ids = tuple(blk.parameters.uuids())
            else:
                param_ids = tuple(map(stringify, blk.parameters.expressions()))
        return param_ids


@dataclass(frozen=True, eq=True)
class Backend(ABC):
    """The abstract class that defines the interface for the backends

    Attributes:
        name: backend unique string identifier
        supports_ad: whether or not the backend has a native autograd
        supports_bp: whether or not the backend has a native backprop
        is_remote: whether computations are executed locally or remotely on this
            backend, useful when using cloud platforms where credentials are
            needed for example.
        with_measurements: whether it supports counts or not
        with_noise: whether to add realistic noise or not
    """

    name: BackendName
    supports_ad: bool
    support_bp: bool
    is_remote: bool
    with_measurements: bool
    native_endianness: Endianness

    # FIXME: should this also go into the configuration?
    with_noise: bool

    # additional configuration specific for each backend
    # some backends might not provide any additional configuration
    # but they will still have an empty Configuration class
    config: BackendConfiguration

    def __post_init__(self) -> None:
        if isinstance(self.config, dict):
            default_conf = self.default_configuration()
            ConfCls = default_conf.__class__

            try:
                new_conf = ConfCls.from_dict(self.config)

                # need this since it is a frozen dataclass
                # see reference documentation
                # https://docs.python.org/3/library/dataclasses.html#frozen-instances
                super().__setattr__("config", new_conf)

            except ValueError as e:
                raise ValueError(f"Wrong configuration provided.\n{str(e)}")

    @abstractmethod
    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        """Converts an abstract `QuantumCircuit` to the native backend representation.

        Arguments:
            circuit: A circuit, for example: `QuantumCircuit(2, X(0))`

        Returns:
            A converted circuit `c`. You can access the original, arbstract circuit via `c.abstract`
            and the converted (or backend *native*) circuit via `c.native`.
        """
        raise NotImplementedError

    @abstractmethod
    def observable(self, observable: AbstractBlock, n_qubits: int) -> ConvertedObservable:
        """Converts an abstract observable (which is just an `AbstractBlock`) to the native backend
        representation.

        Arguments:
            observable: An observable.
            n_qubits: Number of qubits the observable covers. This is typically `circuit.n_qubits`.

        Returns:
            A converted observable `o`. You can access the original, arbstract observable via
            `o.abstract` and the converted (or backend *native*) observable via `o.native`.
        """
        raise NotImplementedError

    def convert(
        self, circuit: QuantumCircuit, observable: list[AbstractBlock] | AbstractBlock | None = None
    ) -> Converted:
        """Convert an abstract circuit (and optionally and observable) to their native
        representation. Additionally this function constructs an embedding function which maps from
        user-facing parameters to device parameters (read more on parameter embedding
        [here][qadence.blocks.embedding.embedding]).
        """

        def check_observable(obs_obj: Any) -> AbstractBlock:
            if isinstance(obs_obj, QubitOperator):
                from qadence.blocks.manipulate import from_openfermion

                assert len(obs_obj.terms) > 0, "Make sure to give a non-empty qubit hamiltonian"

                return from_openfermion(obs_obj)

            elif isinstance(obs_obj, (CompositeBlock, PrimitiveBlock, ScaleBlock)):
                from qadence.blocks.utils import block_is_qubit_hamiltonian

                assert block_is_qubit_hamiltonian(
                    obs_obj
                ), "Make sure the QubitHamiltonian consists only of Pauli operators X, Y, Z, I"
                return obs_obj
            raise TypeError(
                "qubit_hamiltonian should be a Pauli-like AbstractBlock or a QubitOperator"
            )

        conv_circ = self.circuit(circuit)
        circ_params, circ_embedding_fn = embedding(
            conv_circ.abstract.block, self.config._use_gate_params
        )
        params = circ_params
        if observable is not None:
            observable = observable if isinstance(observable, list) else [observable]
            conv_obs = []
            obs_embedding_fn_list = []

            for obs in observable:
                obs = check_observable(obs)
                c_obs = self.observable(obs, max(circuit.n_qubits, obs.n_qubits))
                obs_params, obs_embedding_fn = embedding(
                    c_obs.abstract, self.config._use_gate_params
                )
                params.update(obs_params)
                obs_embedding_fn_list.append(obs_embedding_fn)
                conv_obs.append(c_obs)

            def embedding_fn_dict(a: dict, b: dict) -> dict:
                embedding_dict = circ_embedding_fn(a, b)
                for o in obs_embedding_fn_list:
                    embedding_dict.update(o(a, b))
                return embedding_dict

            return Converted(conv_circ, conv_obs, embedding_fn_dict, params)

        def embedding_fn(a: dict, b: dict) -> dict:
            return circ_embedding_fn(a, b)

        return Converted(conv_circ, None, embedding_fn, params)

    @abstractmethod
    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1000,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        """Sample bit strings.

        Arguments:
            circuit: A converted circuit as returned by `backend.circuit`.
            param_values: _**Already embedded**_ parameters of the circuit. See
                [`embedding`][qadence.blocks.embedding.embedding] for more info.
            n_shots: Number of shots to sample.
            state: Initial state.
            endianness: Endianness of the resulting bitstrings.
        """
        raise NotImplementedError

    @abstractmethod
    def run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """Run a circuit and return the resulting wave function.

        Arguments:
            circuit: A converted circuit as returned by `backend.circuit`.
            param_values: _**Already embedded**_ parameters of the circuit. See
                [`embedding`][qadence.blocks.embedding.embedding] for more info.
            state: Initial state.
            endianness: Endianness of the resulting samples.

        Returns:
            A list of Counter objects where each key represents a bitstring
            and its value the number of times it has been sampled from the given wave function.
        """
        raise NotImplementedError

    @abstractmethod
    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """Compute the expectation value of the `circuit` with the given `observable`.

        Arguments:
            circuit: A converted circuit as returned by `backend.circuit`.
            param_values: _**Already embedded**_ parameters of the circuit. See
                [`embedding`][qadence.blocks.embedding.embedding] for more info.
            state: Initial state.
            endianness: Endianness of the resulting bitstrings.
        """
        raise NotImplementedError

    @abstractmethod
    def assign_parameters(self, circuit: ConvertedCircuit, param_values: dict[str, Tensor]) -> Any:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def default_configuration() -> BackendConfiguration:
        raise NotImplementedError

    def default_diffmode(self) -> DiffMode:
        if self.supports_ad:
            return DiffMode.AD
        else:
            return DiffMode.GPSR


class ConvertedCircuit(Module):
    def __init__(self, native: Any, abstract: QuantumCircuit, original: QuantumCircuit):
        super().__init__()
        self.native = native
        self.abstract = abstract
        self.original = original


class ConvertedObservable(Module):
    def __init__(self, native: Any, abstract: AbstractBlock, original: AbstractBlock):
        super().__init__()
        self.native = native
        self.abstract = abstract
        self.original = original


@dataclass(frozen=True)
class Converted:
    circuit: ConvertedCircuit
    observable: list[ConvertedObservable] | ConvertedObservable | None
    embedding_fn: Callable
    params: dict[str, Tensor]

    def __iter__(self) -> Iterator:
        yield self.circuit
        yield self.observable
        yield self.embedding_fn
        yield self.params



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

from qadence.backend import BackendConfiguration
from qadence.types import AlgoHEvo, Interaction


@dataclass
class Configuration(BackendConfiguration):
    # FIXME: currently not used
    # determine which kind of Hamiltonian evolution
    # algorithm to use
    algo_hevo: AlgoHEvo = AlgoHEvo.EXP

    # number of steps for the Hamiltonian evolution
    n_steps_hevo: int = 100

    use_gradient_checkpointing: bool = False
    """Use gradient checkpointing. Recommended for higher-order optimization tasks."""

    use_single_qubit_composition: bool = False
    """Composes chains of single qubit gates into a single matmul if possible."""

    interaction: Callable | Interaction | str = Interaction.NN
    """Digital-analog emulation interaction that is used for `AnalogBlock`s."""

    loop_expectation: bool = False
    """When computing batches of expectation values, only allocate one wavefunction and loop over
    the batch of parameters to only allocate a single wavefunction at any given time."""



from __future__ import annotations

from functools import partial
from typing import Callable

import torch
from torch import Tensor

from qadence.utils import _round_complex


def general_psr(spectrum: Tensor, n_eqs: int | None = None, shift_prefac: float = 0.5) -> Callable:
    diffs = _round_complex(spectrum - spectrum.reshape(-1, 1))
    sorted_unique_spectral_gaps = torch.unique(torch.abs(torch.tril(diffs)))

    # We have to filter out zeros
    sorted_unique_spectral_gaps = sorted_unique_spectral_gaps[sorted_unique_spectral_gaps > 0]
    n_eqs = len(sorted_unique_spectral_gaps)
    sorted_unique_spectral_gaps = torch.tensor(list(sorted_unique_spectral_gaps))

    if n_eqs == 1:
        return partial(single_gap_psr, spectral_gap=sorted_unique_spectral_gaps.item())
    else:
        return partial(
            multi_gap_psr,
            spectral_gaps=sorted_unique_spectral_gaps,
            shift_prefac=shift_prefac,
        )


def single_gap_psr(
    expectation_fn: Callable[[dict[str, Tensor]], Tensor],
    param_dict: dict[str, Tensor],
    param_name: str,
    spectral_gap: Tensor = torch.tensor([2], dtype=torch.get_default_dtype()),
    shift: Tensor = torch.tensor([torch.pi / 2], dtype=torch.get_default_dtype()),
) -> Tensor:
    """Implements single qubit PSR rule.

    Args:
        expectation_fn (Callable[[dict[str, Tensor]], Tensor]): backend-dependent function
        to calculate expectation value

        param_dict (dict[str, Tensor]): dict storing parameters of parameterized blocks
        param_name (str): name of parameter with respect to that differentiation is performed

    Returns:
        Tensor: tensor containing derivative values
    """

    # + pi/2 shift
    shifted_params = param_dict.copy()
    shifted_params[param_name] = shifted_params[param_name] + shift
    f_plus = expectation_fn(shifted_params)

    # - pi/2 shift
    shifted_params = param_dict.copy()
    shifted_params[param_name] = shifted_params[param_name] - shift
    f_min = expectation_fn(shifted_params)

    return spectral_gap * (f_plus - f_min) / (4 * torch.sin(spectral_gap * shift / 2))


def multi_gap_psr(
    expectation_fn: Callable[[dict[str, Tensor]], Tensor],
    param_dict: dict[str, Tensor],
    param_name: str,
    spectral_gaps: Tensor,
    shift_prefac: float = 0.5,
) -> Tensor:
    """Implements multi-gap multi-qubit GPSR rule.

    Args:
        expectation_fn (Callable[[dict[str, Tensor]], Tensor]): backend-dependent function
        to calculate expectation value

        param_dict (dict[str, Tensor]): dict storing parameters values of parameterized blocks
        param_name (str): name of parameter with respect to that differentiation is performed
        spectral_gaps (Tensor): tensor containing spectral gap values
        shift_prefac (float): prefactor governing the magnitude of parameter shift values -
        select smaller value if spectral gaps are large

    Returns:
        Tensor: tensor containing derivative values
    """
    n_eqs = len(spectral_gaps)
    batch_size = max(t.size(0) for t in param_dict.values())

    # get shift values
    shifts = shift_prefac * torch.linspace(
        torch.pi / 2 - torch.pi / 5, torch.pi / 2 + torch.pi / 5, n_eqs
    )

    # calculate F vector and M matrix
    # (see: https://arxiv.org/pdf/2108.01218.pdf on p. 4 for definitions)
    F = []
    M = torch.empty((n_eqs, n_eqs))
    n_obs = 1
    for i in range(n_eqs):
        # + shift
        shifted_params = param_dict.copy()
        shifted_params[param_name] = shifted_params[param_name] + shifts[i]
        f_plus = expectation_fn(shifted_params)

        # - shift
        shifted_params = param_dict.copy()
        shifted_params[param_name] = shifted_params[param_name] - shifts[i]
        f_minus = expectation_fn(shifted_params)

        F.append((f_plus - f_minus))

        # calculate M matrix
        for j in range(n_eqs):
            M[i, j] = 4 * torch.sin(shifts[i] * spectral_gaps[j] / 2)

    # get number of observables from expectation value tensor
    if f_plus.numel() > 1:
        n_obs = F[0].shape[1]

    # reshape F vector
    F = torch.stack(F).reshape(n_eqs, -1)

    # calculate R vector
    R = torch.linalg.solve(M, F)

    # calculate df/dx
    dfdx = torch.sum(spectral_gaps[:, None] * R, dim=0).reshape(batch_size, n_obs)

    return dfdx



from __future__ import annotations

from collections import Counter
from copy import deepcopy
from functools import singledispatch
from typing import Any, overload

import numpy as np
from torch import Tensor, tensor

from qadence import QuantumCircuit
from qadence.blocks import AbstractBlock


def reassign(block: AbstractBlock, qubit_map: dict[int, int]) -> AbstractBlock:
    """Update the support of a given block

    Args:
        block (AbstractBlock): _description_
        qubit_map (dict[int, int]): _description_

    """
    from qadence.blocks import CompositeBlock, ControlBlock, ParametricControlBlock, ScaleBlock
    from qadence.blocks.utils import _construct

    def _block_with_updated_support(block: AbstractBlock) -> AbstractBlock:
        if isinstance(block, ControlBlock) or isinstance(block, ParametricControlBlock):
            old_qs = block.qubit_support
            new_control_block = deepcopy(block)
            new_control_block._qubit_support = tuple(qubit_map[i] for i in old_qs)
            (subblock,) = block.blocks
            new_control_block.blocks = (reassign(subblock, qubit_map),)  # type: ignore [assignment]
            return new_control_block
        elif isinstance(block, CompositeBlock):
            subblocks = tuple(_block_with_updated_support(b) for b in block.blocks)
            blk = _construct(type(block), subblocks)
            blk.tag = block.tag
            return blk
        elif isinstance(block, ScaleBlock):
            blk = deepcopy(block)  # type: ignore [assignment]
            blk.block = _block_with_updated_support(block.block)  # type: ignore [attr-defined]
            return blk
        else:
            blk = deepcopy(block)  # type: ignore [assignment]
            qs = tuple(qubit_map[i] for i in block.qubit_support)
            blk._qubit_support = qs  # type: ignore[attr-defined]
            return blk

    return _block_with_updated_support(block)


@overload
def invert_endianness(wf: Tensor) -> Tensor:
    ...


@overload
def invert_endianness(arr: np.ndarray) -> np.ndarray:
    ...


@overload
def invert_endianness(cntr: Counter) -> Counter:
    ...


@overload
def invert_endianness(cntrs: list) -> list:
    ...


@overload
def invert_endianness(circuit: QuantumCircuit, n_qubits: int) -> QuantumCircuit:
    ...


@overload
def invert_endianness(block: AbstractBlock, n_qubits: int, in_place: bool) -> AbstractBlock:
    ...


@singledispatch
def invert_endianness(
    x: QuantumCircuit | AbstractBlock | Tensor | Counter | np.ndarray, *args: Any
) -> QuantumCircuit | AbstractBlock | Tensor | Counter | np.ndarray:
    """Invert the endianness of a QuantumCircuit, AbstractBlock, wave function or Counter."""
    raise NotImplementedError(f"Unable to invert endianness of object {type(x)}.")


@invert_endianness.register(AbstractBlock)  # type: ignore[attr-defined]
def _(block: AbstractBlock, n_qubits: int = None, in_place: bool = False) -> AbstractBlock:
    if n_qubits is None:
        n_qubits = block.n_qubits
    """Flips endianness of the block"""
    if in_place:
        raise NotImplementedError
    bits = list(range(n_qubits))
    qubit_map = {i: j for (i, j) in zip(bits, reversed(bits))}
    return reassign(block, qubit_map=qubit_map)


@invert_endianness.register(Tensor)  # type: ignore[attr-defined]
def _(wf: Tensor) -> Tensor:
    """
    Inverts the endianness of a wave function.

    Args:
        wf (Tensor): the target wf as a torch Tensor of shape batch_size X 2**n_qubits

    Returns:
        The inverted wave function.
    """
    n_qubits = int(np.log2(wf.shape[1]))
    ls = list(range(2**n_qubits))
    permute_ind = tensor([int(f"{num:0{n_qubits}b}"[::-1], 2) for num in ls])
    return wf[:, permute_ind]


@invert_endianness.register(np.ndarray)  # type: ignore[attr-defined]
def _(arr: np.ndarray) -> np.ndarray:
    return invert_endianness(tensor(arr)).numpy()


@invert_endianness.register(Counter)  # type: ignore[attr-defined]
def _(cntr: Counter) -> Counter:
    return Counter(
        {
            format(int(bstring[::-1], 2), "0{}b".format(len(bstring))): count
            for bstring, count in cntr.items()
        }
    )


@invert_endianness.register(list)  # type: ignore[attr-defined]
def _(cntrs: list) -> list:
    return list(map(invert_endianness, cntrs))


@invert_endianness.register(QuantumCircuit)  # type: ignore[attr-defined]
def _(circuit: QuantumCircuit) -> QuantumCircuit:
    """This method inverts a circuit "vertically"

    All gates are same but qubit indices are ordered inversely,
    such that bitstrings 00111 become 11100 when measured. Handy for
    big-endian <> little-endian conversion

    Returns:
        QuantumCircuit with endianess switched
    """
    return QuantumCircuit(
        circuit.n_qubits, invert_endianness(circuit.block, circuit.n_qubits, False)
    )



from __future__ import annotations

import re
import shutil
import warnings
from copy import deepcopy
from functools import singledispatch
from tempfile import NamedTemporaryFile
from typing import Any

import sympy

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    CompositeBlock,
    ControlBlock,
    ParametricBlock,
    ParametricControlBlock,
    PrimitiveBlock,
    ScaleBlock,
    chain,
)
from qadence.blocks.analog import ConstantAnalogRotation, WaitBlock
from qadence.circuit import QuantumCircuit
from qadence.models import QuantumModel
from qadence.operations import RX, RY, RZ, SWAP, HamEvo, I
from qadence.transpile.block import fill_identities
from qadence.utils import format_parameter

from .vizbackend import Cluster, QuantumCircuitDiagram

# FIXME: move this to a config/theme?
USE_LATEX_LABELS = False
USE_SUBSCRIPTS = True


def fixed(expr: sympy.Basic) -> bool:
    return expr.is_number  # type: ignore


def trainable(expr: sympy.Basic) -> bool:
    for s in expr.free_symbols:
        if hasattr(s, "trainable") and s.trainable:
            return True
    return False


def _get_latex_label(block: AbstractBlock, color: str = "red", fontsize: int = 30) -> str:
    from latex2svg import default_params, latex2svg
    from lxml import etree

    # FIXME: use this and e.g.
    # qcd.create_node(
    #     wire, label="", image=_get_label(block, qcd.theme.color), block_type="primitive")
    name = sympy.Function(type(block).__name__)
    if isinstance(block, (RX, RY, RZ)):
        p = block.parameters.parameter
        expr = name(p)
    else:
        expr = name

    lx = sympy.latex(expr, mode="inline")

    latex_params = deepcopy(default_params)
    preamble = r"""
    \usepackage[utf8x]{inputenc}
    \usepackage{amsmath}
    \usepackage{amsfonts}
    \usepackage{amssymb}
    \usepackage{amstext}
    \usepackage{xcolor}
    \everymath{\color{black}}
    \everydisplay{\color{black}}
    \def\m@th{\normalcolor\mathsurround\z@}

    """
    preamble = preamble.replace("black", color)
    preamble += r"\usepackage[bitstream-charter]{mathdesign}"
    latex_params["preamble"] = preamble

    svg = latex2svg(lx, params=latex_params)
    svgstr = svg["svg"]
    root = etree.fromstring(svgstr)
    # (x, y_min, width, height) = map(lambda x: fontsize*float(x), root.get("viewBox").split(" "))
    # root.set("viewBox", f"{x_min} {y_min} {width} {height}")
    root.set("width", str(round(svg["width"] * fontsize)))
    root.set("height", str(round(svg["height"] * fontsize)))
    svgstr = etree.tostring(root, encoding="utf-8").decode("utf-8")

    fi = NamedTemporaryFile(delete=False, mode="w", suffix=".svg")
    fi.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>')
    fi.write(svgstr)
    return fi.name


def _is_number(s: str) -> bool:
    # match numbers with an optional minus sign and optional decimal part.
    pattern = re.compile(r"^-?\d+(\.\d+)?$")
    return bool(pattern.match(s))


def _subscript(x: str) -> str:
    offset = ord("₁") - ord("1")
    return chr(ord(x) + offset) if _is_number(x) else x


def _index_to_subscripts(x: str) -> str:
    return re.sub(r"_\d+", lambda match: "".join(map(_subscript, match.group()[1:])), x)


def _expr_string(expr: sympy.Basic) -> str:
    return _index_to_subscripts(format_parameter(expr))


def _get_label(block: AbstractBlock) -> str:
    name = sympy.Function(type(block).__name__)
    if isinstance(block, ParametricBlock):
        p = block.parameters.parameter
        expr = name(p)
    else:
        expr = name
    return _expr_string(expr) if USE_SUBSCRIPTS else format_parameter(expr)


@singledispatch
def make_diagram(
    x: Any,
    qcd: QuantumCircuitDiagram | Cluster | None = None,
    layout: str = "LR",
    theme: str = "light",
    fill: bool = True,
    **kwargs: Any,
) -> QuantumCircuitDiagram:
    raise ValueError(f"Cannot construct circuit diagram for type: {type(x)}")


@make_diagram.register
def _(circuit: QuantumCircuit, *args: Any, **kwargs: Any) -> QuantumCircuitDiagram:
    # FIXME: add register plot here
    return make_diagram(circuit.block, *args, nb_wires=circuit.n_qubits, **kwargs)


@make_diagram.register
def _(model: QuantumModel, *args: Any, **kwargs: Any) -> QuantumCircuitDiagram:
    raise ValueError("Not yet supported.")

    # FIXME: include measurement icon
    if model.out_features > 1:
        raise ValueError("Cannot visualize QuantumModel with more than one observable.")

    obs = deepcopy(model._observable[0].original)
    obs.tag = "Obs."

    block = chain(model._circuit.original.block, obs)
    return make_diagram(block, *args, **kwargs)


@make_diagram.register
def _(
    block: AbstractBlock,
    qcd: QuantumCircuitDiagram = None,
    layout: str = "LR",
    theme: str = "light",
    fill: bool = True,
    **kwargs: Any,
) -> QuantumCircuitDiagram:
    if fill:
        from qadence.transpile import transpile

        block = transpile(
            lambda b: fill_identities(b, 0, b.n_qubits),
            # FIXME: enabling flatten can sometimes prevent wires from bending
            # but flatten currently gets rid of some tags... fix that and comment in:
            # flatten
        )(block)

    if qcd is None:
        nb_wires = kwargs.pop("nb_wires") if "nb_wires" in kwargs else block.n_qubits
        qcd = QuantumCircuitDiagram(nb_wires=nb_wires, layout=layout, theme=theme, **kwargs)

    if isinstance(block, I):
        wire = block.qubit_support[0]
        qcd.create_identity_node(wire)

    elif isinstance(block, SWAP):
        qcd.create_swap_gate(*block.qubit_support)  # type: ignore

    elif isinstance(block, (ControlBlock, ParametricControlBlock)):
        from_wire = block.qubit_support[-1]
        to_wires = block.qubit_support[:-1]
        (b,) = block.blocks
        if isinstance(b, ParametricBlock):
            if block.parameters.parameter.is_number:  # type: ignore[union-attr]
                attrs = qcd.theme.get_fixed_parametric_node_attr()
            elif trainable(block.parameters.parameter):  # type: ignore[union-attr]
                attrs = qcd.theme.get_variational_parametric_node_attr()
            else:
                attrs = qcd.theme.get_feature_parametric_node_attr()
        else:
            attrs = qcd.theme.get_primitive_node_attr()

        qcd.create_control_gate(
            from_wire, to_wires, label=_get_label(b), **attrs  # type: ignore[arg-type]
        )

    elif isinstance(block, HamEvo):
        labels = [block.name, f"t = {_expr_string(block.parameters.parameter)}"]
        start, stop = min(block.qubit_support), block.n_qubits
        _make_cluster(qcd, labels, start, stop, qcd.theme.get_hamevo_cluster_attr())

    elif isinstance(block, AddBlock):
        labels = ["AddBlock"]
        start, stop = min(block.qubit_support), block.n_qubits
        _make_cluster(qcd, labels, start, stop, qcd.theme.get_add_cluster_attr())

    elif isinstance(block, WaitBlock):
        labels = ["wait", f"t = {_expr_string(block.parameters.duration)}"]
        is_global = block.qubit_support.is_global
        start = 0 if is_global else min(block.qubit_support)
        stop = qcd.nb_wires if is_global else block.n_qubits
        _make_cluster(qcd, labels, start, stop, qcd.theme.get_add_cluster_attr())

    elif isinstance(block, ConstantAnalogRotation):
        labels = [
            "AnalogRot",
            f"α = {_expr_string(block.parameters.alpha)}",
            f"t = {_expr_string(block.parameters.duration)}",
            f"Ω = {_expr_string(block.parameters.omega)}",
            f"δ = {_expr_string(block.parameters.delta)}",
            f"φ = {_expr_string(block.parameters.phase)}",
        ]
        is_global = block.qubit_support.is_global
        start = 0 if is_global else min(block.qubit_support)
        stop = qcd.nb_wires if is_global else block.n_qubits
        _make_cluster(qcd, labels, start, stop, qcd.theme.get_add_cluster_attr())

    elif isinstance(block, ScaleBlock):
        s = f"[* {_expr_string(block.scale)}]"
        label = s if block.tag is None else f"{block.tag}: {s}"
        cluster = qcd.create_cluster(label, **qcd.theme.get_scale_cluster_attr())  # type: ignore
        make_diagram(block.block, cluster)

    elif isinstance(block, ParametricBlock):
        wire = block.qubit_support[0]

        if block.parameters.parameter.is_number:
            attrs = qcd.theme.get_fixed_parametric_node_attr()
        elif trainable(block.parameters.parameter):
            attrs = qcd.theme.get_variational_parametric_node_attr()
        else:
            attrs = qcd.theme.get_feature_parametric_node_attr()

        if USE_LATEX_LABELS and shutil.which("latex"):
            qcd.create_node(
                wire,
                label="",
                image=_get_latex_label(block, attrs["color"]),
                **attrs,  # type: ignore[arg-type]
            )
        else:
            if USE_LATEX_LABELS:
                warnings.warn(
                    "To get prettier circuit drawings, consider installing LaTeX.", UserWarning
                )
            qcd.create_node(wire, label=_get_label(block), **attrs)  # type: ignore[arg-type]

    elif isinstance(block, PrimitiveBlock):
        wire = block.qubit_support[0]
        attrs = qcd.theme.get_primitive_node_attr()
        if USE_LATEX_LABELS and shutil.which("latex"):
            qcd.create_node(
                wire,
                label="",
                image=_get_latex_label(block, attrs["color"]),
                **attrs,  # type: ignore[arg-type]
            )
        else:
            if USE_LATEX_LABELS:
                warnings.warn(
                    "To get prettier circuit drawings, consider installing LaTeX.", UserWarning
                )
            qcd.create_node(wire, label=_get_label(block), **attrs)  # type: ignore[arg-type]

    elif isinstance(block, CompositeBlock):
        for inner_block in block:
            if inner_block.tag is not None:
                cluster = qcd.create_cluster(
                    inner_block.tag, **qcd.theme.get_cluster_attr()  # type: ignore
                )
                make_diagram(inner_block, cluster, fill=False)
            else:
                make_diagram(inner_block, qcd, fill=False)

    else:
        raise ValueError(f"Don't know how to draw block of type {type(block)}.")

    return qcd


def _make_cluster(
    qcd: QuantumCircuitDiagram, labels: list[str], start: int, stop: int, attrs: dict
) -> None:
    """Construct a cluster with the list of labels centered vertically (in terms of wires).
    If there are fewer wires than labels, plot all lables in one line, assuming that the first
    element in `labels` is the block type."""
    N = stop - start

    # draw labels line by line
    if N > len(labels):
        cluster = qcd.create_cluster("", show=False, **attrs)
        before = (N - len(labels)) // 2
        after = N - len(labels) - before
        lines = ["" for _ in range(before)] + labels + ["" for _ in range(after)]
        for i, label in zip(range(start, stop), lines):
            _attrs = deepcopy(qcd.theme.get_node_attr())
            _attrs["shape"] = "none"
            _attrs["style"] = "rounded"
            cluster.show = True
            cluster.create_node(i, label=label, **_attrs)
            cluster.show = False

    # draw all labels in one line if there are too few wires
    else:
        cluster = qcd.create_cluster("", show=False, **attrs)
        label = f"{labels[0]}({', '.join(s.replace(' ', '') for s in labels[1:])})"
        for i in range(start, stop):
            if i == ((stop - start) // 2 + start):
                _attrs = deepcopy(qcd.theme.get_node_attr())
                _attrs["shape"] = "none"
                _attrs["style"] = "rounded"
                cluster.show = True
                cluster.create_node(i, label=label, **_attrs)
                cluster.show = False
            else:
                cluster.create_node(i, label="I", **qcd.theme.get_node_attr())



from __future__ import annotations

from collections import Counter
from typing import Any, Callable

import numpy as np
import torch
from torch import Tensor

from qadence.backend import BackendConfiguration, BackendName
from qadence.backends.pytorch_wrapper import DiffMode
from qadence.blocks import AbstractBlock, chain, kron, tag
from qadence.circuit import QuantumCircuit
from qadence.divergences import js_divergence
from qadence.measurements import Measurements
from qadence.models import QuantumModel
from qadence.operations import SWAP, H, I, S, Z
from qadence.transpile import reassign
from qadence.types import OverlapMethod

# Modules to be automatically added to the qadence namespace
__all__ = ["Overlap", "OverlapMethod"]


def _cswap(control: int, target1: int, target2: int) -> AbstractBlock:
    # define projectors on control qubit
    p0 = 0.5 * I(control) + 0.5 * Z(control)
    p1 = 0.5 * I(control) + (-0.5) * Z(control)

    # construct controlled-SWAP block
    cswap_blocks = kron(p0, I(target1), I(target2)) + kron(p1, SWAP(target1, target2))
    cswap = tag(cswap_blocks, f"CSWAP({control}, {target1}, {target2})")

    return cswap


def _controlled_unitary(control: int, unitary_block: AbstractBlock) -> AbstractBlock:
    n_qubits = unitary_block.n_qubits

    # define projectors on control qubit
    p0 = 0.5 * I(control) + 0.5 * Z(control)
    p1 = 0.5 * I(control) + (-0.5) * Z(control)

    # shift qubit support of unitary
    shifted_unitary_block = reassign(unitary_block, {i: control + i + 1 for i in range(n_qubits)})

    # construct controlled-U block
    cu_blocks = kron(p0, *[I(control + i + 1) for i in range(n_qubits)]) + kron(
        p1, shifted_unitary_block
    )
    cu = tag(cu_blocks, f"c-U({control}, {shifted_unitary_block.qubit_support})")

    return cu


def _is_counter_list(lst: list[Counter]) -> bool:
    return all(map(lambda x: isinstance(x, Counter), lst)) and isinstance(lst, list)


def _select_overlap_method(
    method: OverlapMethod,
    backend: BackendName,
    bra_circuit: QuantumCircuit,
    ket_circuit: QuantumCircuit,
) -> tuple[Callable, QuantumCircuit, QuantumCircuit]:
    if method == OverlapMethod.EXACT:
        fn = overlap_exact

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = fn(bras, kets)
            return overlap

    elif method == OverlapMethod.JENSEN_SHANNON:

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = overlap_jensen_shannon(bras, kets)
            return overlap

    elif method == OverlapMethod.COMPUTE_UNCOMPUTE:
        # create a single circuit from bra and ket circuits
        bra_circuit = QuantumCircuit(
            bra_circuit.n_qubits, bra_circuit.block, ket_circuit.block.dagger()
        )
        ket_circuit = None  # type: ignore[assignment]

        def _overlap_fn(  # type: ignore [misc]
            param_values: dict, bra_calc_fn: Callable, bra_state: Tensor | None, *_: Any
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            overlap = overlap_compute_uncompute(bras)
            return overlap

    elif method == OverlapMethod.SWAP_TEST:
        if backend == BackendName.BRAKET:
            raise ValueError("SWAP test method is not supported by the Braket backend.")

        n_qubits = bra_circuit.n_qubits

        # shift qubit support of bra and ket circuit blocks
        shifted_bra_block = reassign(bra_circuit.block, {i: i + 1 for i in range(n_qubits)})
        shifted_ket_block = reassign(
            ket_circuit.block, {i: i + n_qubits + 1 for i in range(n_qubits)}
        )
        ket_circuit = None  # type: ignore[assignment]

        # construct swap test circuit
        state_blocks = kron(shifted_bra_block, shifted_ket_block)
        cswap_blocks = chain(*[_cswap(0, n + 1, n + 1 + n_qubits) for n in range(n_qubits)])
        swap_test_blocks = chain(H(0), state_blocks, cswap_blocks, H(0))
        bra_circuit = QuantumCircuit(2 * n_qubits + 1, swap_test_blocks)

        def _overlap_fn(  # type: ignore [misc]
            param_values: dict, bra_calc_fn: Callable, bra_state: Tensor | None, *_: Any
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            overlap = overlap_swap_test(bras)
            return overlap

    elif method == OverlapMethod.HADAMARD_TEST:
        if backend == BackendName.BRAKET:
            raise ValueError("Hadamard test method is not supported by the Braket backend.")

        n_qubits = bra_circuit.n_qubits

        # construct controlled bra and ket blocks
        c_bra_block = _controlled_unitary(0, bra_circuit.block)
        c_ket_block = _controlled_unitary(0, ket_circuit.block.dagger())

        # construct swap test circuit for Re part
        re_blocks = chain(H(0), c_bra_block, c_ket_block, H(0))
        bra_circuit = QuantumCircuit(n_qubits + 1, re_blocks)

        # construct swap test circuit for Im part
        im_blocks = chain(H(0), c_bra_block, c_ket_block, S(0), H(0))
        ket_circuit = QuantumCircuit(n_qubits + 1, im_blocks)

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = overlap_hadamard_test(bras, kets)
            return overlap

    return _overlap_fn, bra_circuit, ket_circuit


def overlap_exact(bras: Tensor, kets: Tensor) -> Tensor:
    """Calculate overlap using exact quantum mechanical definition.

    Args:
        bras (Tensor): full bra wavefunctions
        kets (Tensor): full ket wavefunctions

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with each ket
    """
    return torch.abs(torch.sum(bras.conj() * kets, dim=1)) ** 2


def fidelity(bras: Tensor, kets: Tensor) -> Tensor:
    return overlap_exact(bras, kets)


def overlap_jensen_shannon(bras: list[Counter], kets: list[Counter]) -> Tensor:
    """Calculate overlap from bitstring counts using Jensen-Shannon divergence method.

    Args:
        bras (list[Counter]): bitstring counts corresponding to bra wavefunctions
        kets (list[Counter]): bitstring counts corresponding to ket wavefunctions

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with each ket
    """
    return 1 - torch.tensor([js_divergence(p, q) for p, q in zip(bras, kets)])


def overlap_compute_uncompute(bras: Tensor | list[Counter]) -> Tensor:
    """Calculate overlap using compute-uncompute method from full wavefunctions or
    bitstring counts.

    Args:
        bras (Tensor | list[Counter]): full bra wavefunctions or bitstring counts

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with zeros ket
    """
    if isinstance(bras, Tensor):
        # calculate exact overlap of full bra wavefunctions with |0> state
        overlap = torch.abs(bras[:, 0]) ** 2

    elif isinstance(bras, list):
        # estimate overlap as the fraction of shots when "0..00" bitstring was observed
        n_qubits = len(list(bras[0].keys())[0])
        n_shots = sum(list(bras[0].values()))
        overlap = torch.tensor([p["0" * n_qubits] / n_shots for p in bras])

    return overlap


def overlap_swap_test(bras: Tensor | list[Counter]) -> Tensor:
    """Calculate overlap using swap test method from full wavefunctions or
    bitstring counts.

    Args:
        bras (Tensor | list[Counter]): full bra wavefunctions or bitstring counts

    Returns:
        Tensor: overlap tensor
    """
    if isinstance(bras, Tensor):
        n_qubits = int(np.log2(bras.shape[1]))

        # define measurement operator  |0><0| x I
        proj_op = torch.tensor([[1.0, 0.0], [0.0, 0.0]])
        ident_op = torch.diag(torch.tensor([1.0 for _ in range(2 ** (n_qubits - 1))]))
        meas_op = torch.kron(proj_op, ident_op).type(torch.complex128)

        # estimate overlap from ancilla qubit measurement
        prob0 = (bras.conj() * torch.matmul(meas_op, bras.t()).t()).sum(dim=1).real

    elif _is_counter_list(bras):
        # estimate overlap as the fraction of shots when 0 was observed on ancilla qubit
        n_qubits = len(list(bras[0].keys())[0])
        n_shots = sum(list(bras[0].values()))
        prob0 = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras
            ]
        )
    else:
        raise TypeError("Incorrect type passed for bras argument.")

    # construct final overlap tensor
    overlap = 2 * prob0 - 1

    return overlap


def overlap_hadamard_test(
    bras_re: Tensor | list[Counter], bras_im: Tensor | list[Counter]
) -> Tensor:
    """Calculate overlap using Hadamard test method from full wavefunctions or
    bitstring counts.

    Args:
        bras_re (Tensor | list[Counter]): full bra wavefunctions or bitstring counts
        for estimation of overlap's real part
        bras_im (Tensor | list[Counter]): full bra wavefunctions or bitstring counts
        for estimation of overlap's imaginary part

    Returns:
        Tensor: overlap tensor
    """
    if isinstance(bras_re, Tensor) and isinstance(bras_im, Tensor):
        n_qubits = int(np.log2(bras_re.shape[1]))

        # define measurement operator  |0><0| x I
        proj_op = torch.tensor([[1.0, 0.0], [0.0, 0.0]])
        ident_op = torch.diag(torch.tensor([1.0 for _ in range(2 ** (n_qubits - 1))]))
        meas_op = torch.kron(proj_op, ident_op).type(torch.complex128)

        # estimate overlap from ancilla qubit measurement
        prob0_re = (bras_re * torch.matmul(meas_op, bras_re.conj().t()).t()).sum(dim=1).real
        prob0_im = (bras_im * torch.matmul(meas_op, bras_im.conj().t()).t()).sum(dim=1).real

    elif _is_counter_list(bras_re) and _is_counter_list(bras_im):
        # estimate overlap as the fraction of shots when 0 was observed on ancilla qubit
        n_qubits = len(list(bras_re[0].keys())[0])
        n_shots = sum(list(bras_re[0].values()))
        prob0_re = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras_re
            ]
        )
        prob0_im = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras_im
            ]
        )
    else:
        raise TypeError("Incorrect types passed for bras_re and kets_re arguments.")

    # construct final overlap tensor
    overlap = (2 * prob0_re - 1) ** 2 + (2 * prob0_im - 1) ** 2

    return overlap


class Overlap(QuantumModel):
    def __init__(
        self,
        bra_circuit: QuantumCircuit,
        ket_circuit: QuantumCircuit,
        backend: BackendName = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        protocol: Measurements | None = None,
        configuration: BackendConfiguration | dict | None = None,
        method: OverlapMethod = OverlapMethod.EXACT,
    ):
        self.backend_name = backend
        self.method = method

        overlap_fn, bra_circuit, ket_circuit = _select_overlap_method(
            method, backend, bra_circuit, ket_circuit
        )
        self.overlap_fn = overlap_fn

        super().__init__(
            bra_circuit,
            backend=backend,
            diff_mode=diff_mode,
            protocol=protocol,
            configuration=configuration,
        )
        self.bra_feat_param_names = set([inp.name for inp in self.inputs])

        if ket_circuit:
            self.ket_model = QuantumModel(
                ket_circuit,
                backend=backend,
                diff_mode=diff_mode,
                protocol=protocol,
                configuration=configuration,
            )
            self.ket_feat_param_names = set([inp.name for inp in self.ket_model.inputs])
        else:
            self.ket_model = None  # type: ignore [assignment]
            self.ket_feat_param_names = set([])

    def _process_param_values(
        self, bra_param_values: dict[str, Tensor], ket_param_values: dict[str, Tensor]
    ) -> dict:
        # we assume that either batch sizes are equal or 0 in case when no user params
        # are present in bra/ket
        bra_param_values = {
            k: v.reshape(-1) if v.shape == () else v for k, v in bra_param_values.items()
        }
        batch_size_bra = (
            len(list(bra_param_values.values())[0]) if len(bra_param_values) != 0 else 0
        )
        ket_param_values = {
            k: v.reshape(-1) if v.shape == () else v for k, v in ket_param_values.items()
        }
        batch_size_ket = (
            len(list(ket_param_values.values())[0]) if len(ket_param_values) != 0 else 0
        )
        new_bra_param_values = bra_param_values.copy()
        new_ket_param_values = ket_param_values.copy()

        # if len(self.bra_feat_param_names) + len(self.ket_feat_param_names) <= 2:

        if len(self.bra_feat_param_names.union(self.ket_feat_param_names)) == 2:
            # extend bra parameter tensors
            for param_name in new_bra_param_values.keys():
                new_bra_param_values[param_name] = new_bra_param_values[param_name].repeat(
                    batch_size_ket
                )

            # extend ket parameter tensors
            for param_name in new_ket_param_values.keys():
                idxs = torch.cat(
                    [
                        torch.ones(batch_size_bra, dtype=torch.int64) * i
                        for i in range(batch_size_ket)
                    ]
                )
                new_ket_param_values[param_name] = new_ket_param_values[param_name][idxs]

            if self.method in [OverlapMethod.EXACT, OverlapMethod.JENSEN_SHANNON]:
                param_values = {"bra": new_bra_param_values, "ket": new_ket_param_values}
            elif self.method in [
                OverlapMethod.COMPUTE_UNCOMPUTE,
                OverlapMethod.SWAP_TEST,
                OverlapMethod.HADAMARD_TEST,
            ]:
                # merge bra and ket param values to simulate all wavefunctions in one batch
                new_bra_param_values.update(new_ket_param_values)
                param_values = {"bra": new_bra_param_values}
                if self.method == OverlapMethod.HADAMARD_TEST:
                    param_values["ket"] = new_bra_param_values

        elif len(self.bra_feat_param_names.union(self.ket_feat_param_names)) < 2:
            if batch_size_bra == batch_size_ket or batch_size_bra == 0 or batch_size_ket == 0:
                param_values = {"bra": bra_param_values, "ket": ket_param_values}
            else:
                raise ValueError("Batch sizes of both bra and ket parameters must be equal.")

        else:
            raise ValueError("Multiple feature parameters for bra/ket are not currently supported.")

        return param_values

    def forward(  # type: ignore [override]
        self,
        bra_param_values: dict[str, Tensor] = {},
        ket_param_values: dict[str, Tensor] = {},
        bra_state: Tensor | None = None,
        ket_state: Tensor | None = None,
        n_shots: int = 0,
    ) -> Tensor:
        # reformat parameters
        param_values = self._process_param_values(bra_param_values, ket_param_values)

        # determine bra and ket calculation functions
        if n_shots == 0:
            bra_calc_fn = getattr(self, "run")
            ket_calc_fn = getattr(self.ket_model, "run", None)
        else:

            def bra_calc_fn(values: dict, state: Tensor) -> Any:
                return getattr(self, "sample")(values, n_shots, state)

            def ket_calc_fn(values: dict, state: Tensor) -> Any:
                return getattr(self.ket_model, "sample", lambda *_: _)(values, n_shots, state)

        # calculate overlap
        overlap = self.overlap_fn(
            param_values, bra_calc_fn, bra_state, ket_calc_fn, ket_state  # type: ignore [arg-type]
        )

        # reshape output if needed
        if len(self.bra_feat_param_names.union(self.ket_feat_param_names)) < 2:
            overlap = overlap[:, None]
        else:
            batch_size_bra = max(len(list(bra_param_values.values())[0]), 1)
            batch_size_ket = max(len(list(ket_param_values.values())[0]), 1)
            overlap = overlap.reshape((batch_size_ket, batch_size_bra)).t()

        return overlap



from __future__ import annotations

from functools import singledispatch
from typing import Any

from torch import Tensor, rand

from qadence.blocks import AbstractBlock, parameters
from qadence.circuit import QuantumCircuit
from qadence.ml_tools.models import TransformedModule
from qadence.models import QNN, QuantumModel
from qadence.parameters import Parameter, stringify


@singledispatch
def rand_featureparameters(
    x: QuantumCircuit | AbstractBlock | QuantumModel | QNN | TransformedModule, *args: Any
) -> dict[str, Tensor]:
    raise NotImplementedError(f"Unable to generate random featureparameters for object {type(x)}.")


@rand_featureparameters.register
def _(block: AbstractBlock, batch_size: int = 1) -> dict[str, Tensor]:
    non_number_params = [p for p in parameters(block) if not p.is_number]
    feat_params: list[Parameter] = [p for p in non_number_params if not p.trainable]
    return {stringify(p): rand(batch_size, requires_grad=False) for p in feat_params}


@rand_featureparameters.register
def _(circuit: QuantumCircuit, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(circuit.block, batch_size)


@rand_featureparameters.register
def _(qm: QuantumModel, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(qm._circuit.abstract, batch_size)


@rand_featureparameters.register
def _(qnn: QNN, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(qnn._circuit.abstract, batch_size)


@rand_featureparameters.register
def _(tm: TransformedModule, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(tm.model, batch_size)



from __future__ import annotations

from collections import Counter

import numpy as np


def shannon_entropy(counter: Counter) -> float:
    return float(-np.sum([count * np.log(count) for count in counter.values()]))


def js_divergence(counter_p: Counter, counter_q: Counter) -> float:
    """
    Compute the Jensen-Shannon divergence between two probability distributions
    represented as Counter objects.
    The JSD is calculated using only the shared keys between the two input Counter objects.

    Args:
        counter_p (Counter): Counter of bitstring counts for probability mass function P.
        counter_q (Counter): Counter of bitstring counts for probability mass function Q.

    Returns:
        float: The Jensen-Shannon divergence between counter_p and counter_q.
    """
    # Normalise counters
    normalisation_p = np.sum([count for count in counter_p.values()])
    normalisation_q = np.sum([count for count in counter_q.values()])
    counter_p = Counter({k: v / normalisation_p for k, v in counter_p.items()})
    counter_q = Counter({k: v / normalisation_q for k, v in counter_q.items()})

    average_proba_counter = counter_p + counter_q
    average_proba_counter = Counter({k: v / 2.0 for k, v in average_proba_counter.items()})
    average_entropy = shannon_entropy(average_proba_counter)

    entropy_p = shannon_entropy(counter_p)
    entropy_q = shannon_entropy(counter_q)
    return float(average_entropy - (entropy_p + entropy_q) / 2.0)



from __future__ import annotations

import inspect
import warnings
from collections.abc import Callable
from math import isclose, pi
from typing import Union

from sympy import Function, acos

from qadence.blocks import AbstractBlock, KronBlock, chain, kron, tag
from qadence.logger import get_logger
from qadence.operations import PHASE, RX, RY, RZ, H
from qadence.parameters import FeatureParameter, Parameter
from qadence.types import BasisSet, ReuploadScaling, TParameter

logger = get_logger(__name__)

ROTATIONS = [RX, RY, RZ, PHASE]
RotationTypes = type[Union[RX, RY, RZ, PHASE]]


def _set_range(fm_type: BasisSet | type[Function] | str) -> tuple[float, float]:
    if fm_type == BasisSet.FOURIER:
        return (0.0, 2 * pi)
    elif fm_type == BasisSet.CHEBYSHEV:
        return (-1.0, 1.0)
    else:
        return (0.0, 1.0)


RS_FUNC_DICT = {
    ReuploadScaling.CONSTANT: lambda i: 1,
    ReuploadScaling.TOWER: lambda i: float(i + 1),
    ReuploadScaling.EXP: lambda i: float(2**i),
}


def feature_map(
    n_qubits: int,
    support: tuple[int, ...] | None = None,
    param: Parameter | str = "phi",
    op: RotationTypes = RX,
    fm_type: BasisSet | type[Function] | str = BasisSet.FOURIER,
    reupload_scaling: ReuploadScaling | Callable | str = ReuploadScaling.CONSTANT,
    feature_range: tuple[float, float] | None = None,
    target_range: tuple[float, float] | None = None,
    multiplier: Parameter | TParameter | None = None,
) -> KronBlock:
    """Construct a feature map of a given type.

    Arguments:
        n_qubits: Number of qubits the feature map covers. Results in `support=range(n_qubits)`.
        support: Puts one feature-encoding rotation gate on every qubit in `support`. n_qubits in
            this case specifies the total overall qubits of the circuit, which may be wider than the
            support itself, but not narrower.
        param: Parameter of the feature map; you can pass a string or Parameter;
            it will be set as non-trainable (FeatureParameter) regardless.
        op: Rotation operation of the feature map; choose from RX, RY, RZ or PHASE.
        fm_type: Basis set for data encoding; choose from `BasisSet.FOURIER` for Fourier
            encoding, or `BasisSet.CHEBYSHEV` for Chebyshev polynomials of the first kind.
        reupload_scaling: how the feature map scales the data that is re-uploaded for each qubit.
            choose from `ReuploadScaling` enumeration or provide your own function with a single
            int as input and int or float as output.
        feature_range: range of data that the input data is assumed to come from.
        target_range: range of data the data encoder assumes as the natural range. For example,
            in Chebyshev polynomials it is (-1, 1), while for Fourier it may be chosen as (0, 2*pi).
        multiplier: overall multiplier; this is useful for reuploading the feature map serially with
            different scalings; can be a number or parameter/expression.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import feature_map, BasisSet, ReuploadScaling

    fm = feature_map(3, fm_type=BasisSet.FOURIER)
    print(f"{fm = }")

    fm = feature_map(3, fm_type=BasisSet.CHEBYSHEV)
    print(f"{fm = }")

    fm = feature_map(3, fm_type=BasisSet.FOURIER, reupload_scaling = ReuploadScaling.TOWER)
    print(f"{fm = }")
    ```
    """

    # Process input
    if support is None:
        support = tuple(range(n_qubits))
    elif len(support) != n_qubits:
        raise ValueError("Wrong qubit support supplied")

    if op not in ROTATIONS:
        raise ValueError(
            f"Operation {op} not supported. "
            f"Please provide one from {[rot.__name__ for rot in ROTATIONS]}."
        )

    # Backwards compatibility
    if fm_type in ("fourier", "chebyshev", "tower"):
        logger.warning(
            "Selecting `fm_type` as 'fourier', 'chebyshev' or 'tower' is deprecated. "
            "Please use the respective enumerations: 'fm_type = BasisSet.FOURIER', "
            "'fm_type = BasisSet.CHEBYSHEV' or 'reupload_scaling = ReuploadScaling.TOWER'."
        )
        if fm_type == "fourier":
            fm_type = BasisSet.FOURIER
        elif fm_type == "chebyshev":
            fm_type = BasisSet.CHEBYSHEV
        elif fm_type == "tower":
            fm_type = BasisSet.CHEBYSHEV
            reupload_scaling = ReuploadScaling.TOWER

    if isinstance(param, Parameter):
        fparam = param
        fparam.trainable = False
    else:
        fparam = FeatureParameter(param)

    # Set feature and target range
    feature_range = _set_range(fm_type) if feature_range is None else feature_range
    target_range = _set_range(fm_type) if target_range is None else target_range

    # Rescale the feature parameter
    scaling = (max(target_range) - min(target_range)) / (max(feature_range) - min(feature_range))
    shift = min(target_range) - min(feature_range) * scaling

    if isclose(scaling, 1.0):
        # So we don't get 1.0 factor in visualization
        scaled_fparam = fparam + shift
    else:
        scaled_fparam = scaling * fparam + shift

    # Transform feature parameter
    if fm_type == BasisSet.FOURIER:
        transformed_feature = scaled_fparam
    elif fm_type == BasisSet.CHEBYSHEV:
        transformed_feature = acos(scaled_fparam)
    elif inspect.isclass(fm_type) and issubclass(fm_type, Function):
        transformed_feature = fm_type(scaled_fparam)
    else:
        raise NotImplementedError(
            f"Feature map type {fm_type} not implemented. Choose an item from the BasisSet "
            f"enum: {[bs.name for bs in BasisSet]}, or your own sympy.Function to wrap "
            "the given feature parameter with."
        )

    basis_tag = fm_type.value if isinstance(fm_type, BasisSet) else str(fm_type)

    # Set reupload scaling function
    if callable(reupload_scaling):
        rs_func = reupload_scaling
        rs_tag = "Custom"
    else:
        rs_func = RS_FUNC_DICT.get(reupload_scaling, None)  # type: ignore [call-overload]
        if rs_func is None:
            raise NotImplementedError(
                f"Reupload scaling {reupload_scaling} not implemented; choose an item from "
                f"the ReuploadScaling enum: {[rs.name for rs in ReuploadScaling]}, or your own "
                "python function with a single int arg as input and int or float output."
            )
        if isinstance(reupload_scaling, ReuploadScaling):
            rs_tag = reupload_scaling.value
        else:
            rs_tag = reupload_scaling

    # Set overall multiplier
    multiplier = 1 if multiplier is None else multiplier

    # Build feature map
    op_list = []
    for i, qubit in enumerate(support):
        op_list.append(op(qubit, multiplier * rs_func(i) * transformed_feature))
    fm = kron(*op_list)

    fm.tag = rs_tag + " " + basis_tag + " FM"

    return fm


def fourier_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: RotationTypes = RX
) -> AbstractBlock:
    """Construct a Fourier feature map.

    Args:
        n_qubits: number of qubits across which the FM is created
        param: The base name for the feature `Parameter`
    """
    warnings.warn(
        "Function 'fourier_feature_map' is deprecated. Please use 'feature_map' directly.",
        FutureWarning,
    )
    fm = feature_map(n_qubits, support=support, param=param, op=op, fm_type=BasisSet.FOURIER)
    return fm


def chebyshev_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: RotationTypes = RX
) -> AbstractBlock:
    """Construct a Chebyshev feature map.

    Args:
        n_qubits: number of qubits across which the FM is created
        support (Iterable[int]): The qubit support
        param: The base name for the feature `Parameter`
    """
    warnings.warn(
        "Function 'chebyshev_feature_map' is deprecated. Please use 'feature_map' directly.",
        FutureWarning,
    )
    fm = feature_map(n_qubits, support=support, param=param, op=op, fm_type=BasisSet.CHEBYSHEV)
    return fm


def tower_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: RotationTypes = RX
) -> AbstractBlock:
    """Construct a Chebyshev tower feature map.

    Args:
        n_qubits: number of qubits across which the FM is created
        param: The base name for the feature `Parameter`
    """
    warnings.warn(
        "Function 'tower_feature_map' is deprecated. Please use feature_map directly.",
        FutureWarning,
    )
    fm = feature_map(
        n_qubits,
        support=support,
        param=param,
        op=op,
        fm_type=BasisSet.CHEBYSHEV,
        reupload_scaling=ReuploadScaling.TOWER,
    )
    return fm


def exp_fourier_feature_map(
    n_qubits: int,
    support: tuple[int, ...] = None,
    param: str = "x",
    feature_range: tuple[float, float] = None,
) -> AbstractBlock:
    """
    Exponential fourier feature map.

    Args:
        n_qubits: number of qubits in the feature
        support: qubit support
        param: name of feature `Parameter`
        feature_range: min and max value of the feature, as floats in a Tuple
    """

    if feature_range is None:
        feature_range = (0.0, 2.0**n_qubits)

    support = tuple(range(n_qubits)) if support is None else support
    hlayer = kron(H(qubit) for qubit in support)
    rlayer = feature_map(
        n_qubits,
        support=support,
        param=param,
        op=RZ,
        fm_type=BasisSet.FOURIER,
        reupload_scaling=ReuploadScaling.EXP,
        feature_range=feature_range,
        target_range=(0.0, 2 * pi),
    )
    rlayer.tag = None
    return tag(chain(hlayer, rlayer), f"ExpFourierFM({param})")



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from typing import Any

import numpy as np
import qutip
import torch
from pulser import Register as PulserRegister
from pulser import Sequence
from pulser.pulse import Pulse
from pulser_simulation.simresults import SimulationResults
from pulser_simulation.simulation import QutipEmulator
from torch import Tensor

from qadence.backend import Backend as BackendInterface
from qadence.backend import BackendName, ConvertedCircuit, ConvertedObservable
from qadence.backends.utils import to_list_of_dicts
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.overlap import overlap_exact
from qadence.register import Register
from qadence.utils import Endianness

from .channels import GLOBAL_CHANNEL, LOCAL_CHANNEL
from .config import Configuration
from .convert_ops import convert_observable
from .devices import Device, IdealDevice, RealisticDevice
from .pulses import add_pulses

WEAK_COUPLING_CONST = 1.2

DEFAULT_SPACING = 8.0  # µm (standard value)


def create_register(register: Register, spacing: float = DEFAULT_SPACING) -> PulserRegister:
    """Create Pulser register instance.

    Args:
        register (Register): graph representing a register with accompanying coordinate data
        spacing (float): distance between qubits in micrometers

    Returns:
        Register: Pulser register
    """

    # create register from coordinates
    coords = np.array(list(register.coords.values()))
    return PulserRegister.from_coordinates(coords * spacing)


def make_sequence(circ: QuantumCircuit, config: Configuration) -> Sequence:
    if config.device_type == Device.IDEALIZED:
        device = IdealDevice
    elif config.device_type == Device.REALISTIC:
        device = RealisticDevice
    else:
        raise ValueError("Specified device is not supported.")

    max_amp = device.channels["rydberg_global"].max_amp
    min_duration = device.channels["rydberg_global"].min_duration

    if config.spacing is not None:
        spacing = config.spacing
    elif max_amp is not None:
        # Ideal spacing for entanglement gate
        spacing = WEAK_COUPLING_CONST * device.rydberg_blockade_radius(max_amp)  # type: ignore
    else:
        spacing = DEFAULT_SPACING

    pulser_register = create_register(circ.register, spacing)

    sequence = Sequence(pulser_register, device)
    sequence.declare_channel(GLOBAL_CHANNEL, "rydberg_global")
    sequence.declare_channel(LOCAL_CHANNEL, "rydberg_local", initial_target=0)

    # add a minimum duration pulse omega=0 pulse at the beginning for simulation convergence reasons
    # since Pulser's QutipEmulator doesn't allow simulation of sequences with total duration < 4ns
    zero_pulse = Pulse.ConstantPulse(
        duration=max(sequence.device.channels["rydberg_global"].min_duration, 4),
        amplitude=0.0,
        detuning=0.0,
        phase=0.0,
    )
    sequence.add(zero_pulse, GLOBAL_CHANNEL, "wait-for-all")

    add_pulses(sequence, circ.block, config, circ.register, spacing)
    sequence.measure()

    return sequence


# TODO: make it parallelized
# TODO: add execution on the cloud platform
def simulate_sequence(
    sequence: Sequence, config: Configuration, state: Tensor
) -> SimulationResults:
    simulation = QutipEmulator.from_sequence(
        sequence,
        sampling_rate=config.sampling_rate,
        config=config.sim_config,
        with_modulation=config.with_modulation,
    )
    if state is not None:
        simulation.set_initial_state(qutip.Qobj(state.cpu().numpy()))

    return simulation.run(nsteps=config.n_steps_solv, method=config.method_solv)


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    """The Pulser backend"""

    name: BackendName = BackendName.PULSER
    supports_ad: bool = False
    support_bp: bool = False
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = Configuration()

    def circuit(self, circ: QuantumCircuit) -> Sequence:
        native = make_sequence(circ, self.config)

        return ConvertedCircuit(native=native, abstract=circ, original=circ)

    def observable(self, observable: AbstractBlock, n_qubits: int = None) -> Tensor:
        from qadence.transpile import flatten, scale_primitive_blocks_only, transpile

        # make sure only leaves, i.e. primitive blocks are scaled
        block = transpile(flatten, scale_primitive_blocks_only)(observable)

        (native,) = convert_observable(block, n_qubits=n_qubits, config=self.config)
        return ConvertedObservable(native=native, abstract=block, original=observable)

    def assign_parameters(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor],
    ) -> Any:
        if param_values == {} and circuit.native.is_parametrized():
            missing = list(circuit.native.declared_variables.keys())
            raise ValueError(f"Please, provide values for the following parameters: {missing}")

        if param_values == {}:
            return circuit.native

        numpy_param_values = {
            k: v.detach().cpu().numpy()
            for (k, v) in param_values.items()
            if k in circuit.native.declared_variables
        }

        return circuit.native.build(**numpy_param_values)

    def run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        vals = to_list_of_dicts(param_values)

        batched_wf = np.zeros((len(vals), 2**circuit.abstract.n_qubits), dtype=np.complex128)

        for i, param_values_el in enumerate(vals):
            sequence = self.assign_parameters(circuit, param_values_el)
            sim_result = simulate_sequence(sequence, self.config, state)
            wf = (
                sim_result.get_final_state(ignore_global_phase=False, normalize=True)
                .full()
                .flatten()
            )

            # We flip the wavefunction coming out of pulser,
            # essentially changing logic 0 with logic 1 in the basis states.
            batched_wf[i] = np.flip(wf)

        batched_wf_torch = torch.from_numpy(batched_wf)

        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            batched_wf_torch = invert_endianness(batched_wf_torch)

        return batched_wf_torch

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        vals = to_list_of_dicts(param_values)

        samples = []
        for param_values_el in vals:
            sequence = self.assign_parameters(circuit, param_values_el)
            sim_result = simulate_sequence(sequence, self.config, state)
            sample = sim_result.sample_final_state(n_shots)
            samples.append(sample)
        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            samples = invert_endianness(samples)
        return samples

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        state = self.run(circuit, param_values=param_values, state=state, endianness=endianness)

        observables = observable if isinstance(observable, list) else [observable]
        support = sorted(list(circuit.abstract.register.support))
        res_list = [obs.native(state, param_values, qubit_support=support) for obs in observables]

        res = torch.transpose(torch.stack(res_list), 0, 1)
        res = res if len(res.shape) > 0 else res.reshape(1)
        return res.real

    @staticmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        return overlap_exact(bras, kets)

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()



from __future__ import annotations

from dataclasses import dataclass

import torch
from torch.utils.data import DataLoader, TensorDataset


@dataclass
class DictDataLoader:
    """This class only holds a dictionary of `DataLoader`s and samples from them"""

    dataloaders: dict[str, DataLoader]

    # this flag indicates that the dictionary contains dataloaders
    # which can automatically iterate at each epoch without having to
    # redefine the iterator itself (so basically no StopIteration exception
    # will occur). This is the case of the Flow library where the dataloader
    # is actually mostly used, so it is set to True by default
    has_automatic_iter: bool = True

    def __iter__(self) -> DictDataLoader:
        self.iters = {key: iter(dl) for key, dl in self.dataloaders.items()}
        return self

    def __next__(self) -> dict[str, torch.Tensor]:
        return {key: next(it) for key, it in self.iters.items()}


def to_dataloader(x: torch.Tensor, y: torch.Tensor, batch_size: int = 1) -> DataLoader:
    """Convert two torch tensors x and y to a Dataloader."""
    return DataLoader(TensorDataset(x, y), batch_size=batch_size)



from __future__ import annotations

from numpy import pi
from pulser.channels.channels import Rydberg
from pulser.channels.eom import RydbergBeam, RydbergEOM
from pulser.devices._device_datacls import Device as PulserDevice
from pulser.devices._device_datacls import VirtualDevice

from qadence.types import StrEnum

# Idealized virtual device
IdealDevice = VirtualDevice(
    name="IdealizedDevice",
    dimensions=2,
    rydberg_level=60,
    max_atom_num=100,
    max_radial_distance=100,
    min_atom_distance=0,
    channel_objects=(
        Rydberg.Global(max_abs_detuning=2 * pi * 4, max_amp=2 * pi * 3),
        Rydberg.Local(max_targets=1000, max_abs_detuning=2 * pi * 4, max_amp=2 * pi * 3),
    ),
)


# device with realistic specs with local channels and custom bandwith.
RealisticDevice = PulserDevice(
    name="RealisticDevice",
    dimensions=2,
    rydberg_level=60,
    max_atom_num=100,
    max_radial_distance=60,
    min_atom_distance=5,
    channel_objects=(
        Rydberg.Global(
            max_abs_detuning=2 * pi * 4,
            max_amp=2 * pi * 1.5,
            clock_period=4,
            min_duration=16,
            max_duration=4000,
            mod_bandwidth=16,
            eom_config=RydbergEOM(
                limiting_beam=RydbergBeam.RED,
                max_limiting_amp=40 * 2 * pi,
                intermediate_detuning=700 * 2 * pi,
                mod_bandwidth=24,
                controlled_beams=(RydbergBeam.BLUE,),
            ),
        ),
        Rydberg.Local(
            max_targets=20,
            max_abs_detuning=2 * pi * 4,
            max_amp=2 * pi * 3,
            clock_period=4,
            min_duration=16,
            max_duration=2**26,
            mod_bandwidth=16,
            eom_config=RydbergEOM(
                limiting_beam=RydbergBeam.RED,
                max_limiting_amp=40 * 2 * pi,
                intermediate_detuning=700 * 2 * pi,
                mod_bandwidth=24,
                controlled_beams=(RydbergBeam.BLUE,),
            ),
        ),
    ),
)


class Device(StrEnum):
    """Supported types of devices for Pulser backend"""

    IDEALIZED = IdealDevice
    "idealized device, least realistic"

    REALISTIC = RealisticDevice
    "device with realistic specs"



from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Any

import torch
from nevergrad.optimization.base import Optimizer as NGOptimizer
from torch.nn import Module
from torch.optim import Optimizer

from qadence.logger import get_logger

logger = get_logger(__name__)


def get_latest_checkpoint_name(folder: Path, type: str) -> Path:
    file = Path("")
    files = [f for f in os.listdir(folder) if f.endswith(".pt") and type in f]
    if len(files) == 0:
        logger.error(f"Directory {folder} does not contain any {type} checkpoints.")
    if len(files) == 1:
        file = Path(files[0])
    else:
        pattern = re.compile(".*_(\d+).pt$")
        max_index = -1
        for f in files:
            match = pattern.search(f)
            if match:
                index_str = match.group(1).replace("_", "")
                index = int(index_str)
                if index > max_index:
                    max_index = index
                    file = Path(f)
    return Path(file)


def load_checkpoint(
    folder: Path,
    model: Module,
    optimizer: Optimizer | NGOptimizer,
    model_ckpt_name: str | Path = "",
    opt_ckpt_name: str | Path = "",
) -> tuple[Module, Optimizer | NGOptimizer, int]:
    if isinstance(folder, str):
        folder = Path(folder)
    if not folder.exists():
        folder.mkdir(parents=True)
        return model, optimizer, 0
    model, iter = load_model(folder, model, model_ckpt_name)
    optimizer = load_optimizer(folder, optimizer, opt_ckpt_name)
    return model, optimizer, iter


def write_checkpoint(
    folder: Path, model: Module, optimizer: Optimizer | NGOptimizer, iteration: int
) -> None:
    from qadence.ml_tools.models import TransformedModule
    from qadence.models import QNN, QuantumModel

    model_checkpoint_name: str = f"model_{type(model).__name__}_ckpt_" + f"{iteration:03n}" + ".pt"
    opt_checkpoint_name: str = f"opt_{type(optimizer).__name__}_ckpt_" + f"{iteration:03n}" + ".pt"
    try:
        d = (
            model._to_dict(save_params=True)
            if isinstance(model, (QNN, QuantumModel)) or isinstance(model, TransformedModule)
            else model.state_dict()
        )
        torch.save((iteration, d), folder / model_checkpoint_name)
        logger.info(f"Writing {type(model).__name__} checkpoint {model_checkpoint_name}")
    except Exception as e:
        logger.exception(e)
    try:
        if isinstance(optimizer, Optimizer):
            torch.save(
                (iteration, type(optimizer), optimizer.state_dict()), folder / opt_checkpoint_name
            )
        elif isinstance(optimizer, NGOptimizer):
            optimizer.dump(folder / opt_checkpoint_name)
        logger.info(f"Writing {type(optimizer).__name__} to checkpoint {opt_checkpoint_name}")
    except Exception as e:
        logger.exception(e)


def load_model(
    folder: Path, model: Module, model_ckpt_name: str | Path = "", *args: Any, **kwargs: Any
) -> tuple[Module, int]:
    from qadence.ml_tools.models import TransformedModule
    from qadence.models import QNN, QuantumModel

    iteration = 0
    if model_ckpt_name == "":
        model_ckpt_name = get_latest_checkpoint_name(folder, "model")

    try:
        iteration, model_dict = torch.load(folder / model_ckpt_name, *args, **kwargs)
        if isinstance(model, (QuantumModel, QNN, TransformedModule)):
            model._from_dict(model_dict, as_torch=True)
        elif isinstance(model, Module):
            model.load_state_dict(model_dict, strict=True)

    except Exception as e:
        msg = f"Unable to load state dict due to {e}.\
               No corresponding pre-trained model found. Returning the un-trained model."
        import warnings

        warnings.warn(msg, UserWarning)
        logger.warn(msg)
    return model, iteration


def load_optimizer(
    folder: Path,
    optimizer: Optimizer | NGOptimizer,
    opt_ckpt_name: str | Path = "",
) -> Optimizer | NGOptimizer:
    if opt_ckpt_name == "":
        opt_ckpt_name = get_latest_checkpoint_name(folder, "opt")
    if os.path.isfile(folder / opt_ckpt_name):
        if isinstance(optimizer, Optimizer):
            (_, OptType, optimizer_state) = torch.load(folder / opt_ckpt_name)
            if isinstance(optimizer, OptType):
                optimizer.load_state_dict(optimizer_state)

        elif isinstance(optimizer, NGOptimizer):
            loaded_optimizer = NGOptimizer.load(folder / opt_ckpt_name)
            if loaded_optimizer.name == optimizer.name:
                optimizer = loaded_optimizer
        else:
            raise NotImplementedError
    return optimizer



from __future__ import annotations

from dataclasses import dataclass

from pulser.channels.channels import Rydberg

GLOBAL_CHANNEL = "Global"
LOCAL_CHANNEL = "Local"


@dataclass(frozen=True)
class CustomRydberg(Rydberg):
    name: str = "Rydberg"

    duration_steps: int = 1  # ns
    amplitude_steps: float = 0.01  # rad/µs



from __future__ import annotations

from dataclasses import dataclass, field

from qadence.backend import BackendConfiguration


@dataclass
class Configuration(BackendConfiguration):
    # FIXME: currently not used
    # credentials for connecting to the cloud
    # and executing on real QPUs
    cloud_credentials: dict = field(default_factory=dict)
    # Braket requires gate-level parameters
    use_gate_params = True
